from functools import cached_property
from typing import Literal, Sequence

import numpy as np

from .utils import conf


class Fragment:
    """Contains information about a collection of blobs that belong to the
    same animal or to the same crossing."""

    acceptable_for_training: bool | None
    """Boolean to indicate that the fragment was identified sufficiently
    well and can in principle be used for training. See also the
    accumulation_manager.py module."""

    temporary_id: int | None
    """Integer indicating a temporary identity assigned to the fragment
    during the cascade of training and identification protocols."""

    is_certain: bool | None = None
    """Boolean indicating whether the fragment is certain enough to be
    accumulated. See also the accumulation_manager.py module."""

    accumulable: bool | None = None
    """Boolean indicating whether the fragment can be accumulated, i.e. it
    can potentially be used for training."""

    is_in_a_global_fragment: bool = False
    """Indicates whether the fragment is part of a global fragment"""

    P1_vector: np.ndarray
    """Numpy array indicating the P1 probability of each of the possible
    identities"""

    certainty: float
    """Indicates the certainty of the identity"""

    P2_vector: np.ndarray | None = None
    """Numpy array indicating the P2 probability of each of the possible
    identities. See also :meth:`compute_P2_vector`"""

    identity: int | None = None
    """Identity assigned to the fragment during the cascade of training
    and identification protocols or during the residual identification
    (see also the assigner.py module)"""

    non_consistent: bool = False
    """Boolean indicating whether the fragment identity is consistent with
    coexisting fragment"""

    ambiguous_identities: np.ndarray | None = None
    """Identities that would be ambiguously assigned during the residual
    identification process. See also the assigner.py module"""

    used_for_training: bool = False
    """Boolean indicating whether the images in the fragment were used to
    train the identification network during the cascade of training and
    identification protocols. See also the accumulation_manager.py module.
    """
    accumulation_step: int | None = None
    """Integer indicating the accumulation step at which the fragment was
    accumulated. See also the accumulation_manager.py module."""

    identities_corrected_closing_gaps: list[int] | None = None
    """Identity of the fragment assigned during the interpolation of the
        gaps produced by the crossing fragments. See also the
        assign_them_all.py module."""

    identity_corrected_solving_jumps: int | None = None
    """Identity of the fragment assigned during the correction of imposible
    (unrealistic) velocity jumps in the trajectories. See also the
    correct_impossible_velocity_jumps.py module."""

    identity_is_fixed: bool = False
    """Boolean indicating whether the identity is fixed and cannot be
    modified during the postprocessing. This attribute is given during
    the residual identification (see assigner.py module)"""

    P1_below_random: bool | None

    used_for_pretraining = False
    """Boolean indicating whether the images in the fragment were used to
    pretrain the identification network during the pretraining step of the
    Protocol 3. See also the accumulation_manager.py module."""

    accumulated_globally: bool = False
    """Boolean indicating whether the fragment was accumulated in a
    global accumulation step of the cascade of training and identification
    protocols. See also the accumulation_manager.py module."""

    accumulated_partially = False
    """Boolean indicating whether the fragment was accumulated in a
    partial accumulation step of the cascade of training and identification
    protocols. See also the accumulation_manager.py module."""

    user_generated_identity: int | None = None
    """This property is give during the correction of impossible velocity
    jumps. It has nothing to do with the manual validation."""

    coexisting_individual_fragments: list["Fragment"]
    """list of fragment objects representing and individual (i.e.
    not representing a crossing where two or more animals are touching) and
    coexisting (in frame) with self"""

    forced_crossing: bool = False
    """Indicates if the crossing attribute has been forced by set_individual_with_identity_0_as_crossings()"""

    frame_by_frame_velocity: np.ndarray
    """Instant speed (in each frame) of the blob in the fragment"""

    start_position: tuple[float, float]
    """X and Y position of the blob's centroid at the start of the fragment"""

    end_position: tuple[float, float]
    """X and Y position of the blob's centroid at the end of the fragment"""

    def __init__(
        self,
        fragment_identifier: int,
        start_frame: int,
        end_frame: int,
        images: list[int],
        centroids: list[tuple[float, float]],
        episodes: list[int],
        is_an_individual: bool,
    ):
        self.identifier = fragment_identifier
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.images = images
        self.episodes = episodes
        self.is_an_individual = is_an_individual

        if len(centroids) > 1:
            self.frame_by_frame_velocity = np.sqrt(
                (np.diff(centroids, axis=0) ** 2).sum(axis=1)
            )
        else:
            self.frame_by_frame_velocity = np.array([0])

        self.start_position = centroids[0]
        self.end_position = centroids[-1]

    @property
    def image_locations(self):
        return zip(self.images, self.episodes)

    @classmethod
    def from_json(cls, json: dict):
        fragment: cls = cls.__new__(cls)
        fragment.__dict__ = json
        if len(fragment.episodes) == 1:  # decompress
            fragment.episodes = [fragment.episodes[0]] * len(fragment.images)
        for key in (
            "P1_vector",
            "P2_vector",
            "ambiguous_identities",
            "frame_by_frame_velocity",
        ):
            if key in json:
                setattr(fragment, key, np.asarray(json[key]))
        return fragment

    @property
    def distance_travelled(self) -> float:
        """The distance traveled by the individual in the fragment.
        It is based on the position of the centroids in consecutive images.
        """
        return self.frame_by_frame_velocity.sum()

    def reset(
        self,
        roll_back_to: Literal["fragmentation", "accumulation"],
        number_of_animals: int,
    ):
        """Reset attributes of the fragment to a specific part of the
        algorithm.

        Parameters
        ----------
        roll_back_to : str
            Reset all the attributes up to the process specified in input.
            'fragmentation', 'pretraining', 'accumulation', 'assignment'
        """
        #  This method was mainly used to resume the tracking from different
        # rocessing steps. Currently this function is not active, but this
        #  method might still be useful in the future.
        self.identity_is_fixed = False
        if roll_back_to == "fragmentation":
            self.used_for_training = False
            self.used_for_pretraining = False
            self.acceptable_for_training = None
            self.temporary_id = None
            self.identity = None
            self.identity_corrected_solving_jumps = None
            self.accumulated_globally = False
            self.accumulated_partially = False
            self.accumulation_step = None
            self.is_certain = None
            self.non_consistent = False
            self.certainty = 0.0
            self.P1_vector = np.zeros(number_of_animals)
            self.P1_below_random = None
        elif roll_back_to == "accumulation":
            if not self.used_for_training:
                self.identity = None
                self.identity_corrected_solving_jumps = None
                self.P1_vector = np.zeros(number_of_animals)
            self.ambiguous_identities = None
            self.P2_vector = None
        else:
            raise ValueError(roll_back_to)

    @property
    def is_a_crossing(self) -> bool:
        return not self.is_an_individual

    @property
    def assigned_identities(self):
        """Assigned identities (list) by the algorithm considering the
        identification process and the postprocessing steps (correction of
        impossible velocity jumps and interpolation of crossings).

        The fragment can have multiple identities if it is a crossing fragment.
        """
        if self.identities_corrected_closing_gaps is not None:
            return self.identities_corrected_closing_gaps
        if self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        return [self.identity]

    @cached_property
    def number_of_images(self):
        """Number images (or blobs) in the fragment."""
        return len(self.images)

    @property
    def has_enough_accumulated_coexisting_fragments(self):
        """Boolean indicating whether the fragment has enough coexisting and
        already accumulated fragments.

        This property is used during the partial accumulation. See also the
        accumulation_manager.py module.
        """
        return (
            sum(
                fragment.used_for_training
                for fragment in self.coexisting_individual_fragments
            )
            >= len(self.coexisting_individual_fragments) / 2
        )

    def compute_border_velocity(self, other: "Fragment") -> float:
        """Velocity necessary to cover the space between two fragments.

        Note that these velocities are divided by the number of frames that
        separate self and other fragment.

        Parameters
        ----------
        other : :class:`Fragment`
            Another fragment

        Returns
        -------
        float
            Returns the speed at which an individual should travel to be
            present in both self and other fragments.

        """
        if self.start_frame > other.end_frame:
            centroids = np.asarray([self.start_position, other.end_position])
        else:
            centroids = np.asarray([self.end_position, other.start_position])
        return np.sqrt((np.diff(centroids, axis=0) ** 2).sum(axis=1))[0]

    def coexist_with(self, other: "Fragment"):
        """Boolean indicating whether the given fragment coexists in time with
        another fragment.

        Parameters
        ----------
        other :  :class:`Fragment`
            A second fragment

        Returns
        -------
        bool
            True if self and other coexist in time in at least one frame.

        """
        return self.start_frame < other.end_frame and self.end_frame > other.start_frame

    def check_consistency_with_coexistent_individual_fragments(self, temporary_id):
        """Check that the temporary identity assigned to the fragment is
        consistent with respect to the identities already assigned to the
        fragments coexisting (in frame) with it.

        Parameters
        ----------
        temporary_id : int
            Temporary identity assigned to the fragment.

        Returns
        -------
        bool
            True if the identification of self with `temporary_id` does not
            cause any duplication of identities.

        """
        return all(
            coexisting_fragment.temporary_id != temporary_id
            for coexisting_fragment in self.coexisting_individual_fragments
        )

    def compute_identification_statistics(
        self,
        predictions: np.ndarray | list,
        softmax_probs: np.ndarray,
        number_of_animals: int,
    ):
        """Computes the statistics necessary for the identification of the
        fragment.

        Parameters
        ----------
        predictions : numpy array
            Array of shape [number_of_images_in_fragment, 1] whose components
            are the argmax(softmax_probs) per image
        softmax_probs : numpy array
            Array of shape [number_of_images_in_fragment, number_of_animals]
            whose rows are the result of applying the softmax function to the
            predictions outputted by the idCNN per image
        number_of_animals : int
            Description of parameter `number_of_animals`.

        See Also
        --------
        :meth:`compute_median_softmax`
        """
        assert self.is_an_individual

        frequencies = np.bincount(predictions, minlength=number_of_animals + 1)[1:]
        self.set_P1_from_frequencies(frequencies)
        median_softmax = self.compute_median_softmax(softmax_probs, number_of_animals)
        self.set_certainty_of_individual_fragment(median_softmax)

    def assign_identity(self, number_of_animals: int):
        """Assigns the identity to the fragment by considering the fragments
        coexisting with it.

        If the certainty of the identification is high enough it sets
        the identity of the fragment as fixed and it won't be modified during
        the postprocessing.
        """
        assert self.is_an_individual
        if self.identity_is_fixed:
            return
        if self.used_for_training:
            self.identity_is_fixed = True
            return

        assert self.P2_vector is not None

        max_P2 = self.P2_vector.max()  # there can be two equal maximums
        possible_identities = np.argwhere(self.P2_vector == max_P2)[:, 0] + 1

        if len(possible_identities) > 1:  # TODO is it possible?
            self.identity = 0
            self.zero_identity_assigned_by_P2 = True
            self.ambiguous_identities = possible_identities
        else:
            if max_P2 > conf.FIXED_IDENTITY_THRESHOLD:
                self.identity_is_fixed = True
            self.identity = possible_identities[0]
            self.P1_vector = np.zeros(len(self.P1_vector))
            self.P1_vector[self.identity - 1] = 1.0
            for fragment in self.coexisting_individual_fragments:
                fragment.compute_P2_vector(number_of_animals)

    def compute_P2_vector(self, number_of_animals: int):
        """Computes the P2_vector of the fragment.

        It is based on :attr:`coexisting_individual_fragments`"""
        coexisting_P1_vectors = np.asarray(
            [fragment.P1_vector for fragment in self.coexisting_individual_fragments]
        )
        numerator = self.P1_vector * np.prod(1.0 - coexisting_P1_vectors, axis=0)
        denominator = numerator.sum()
        if denominator != 0:
            self.P2_vector = numerator / denominator
        else:
            self.P2_vector = np.zeros(number_of_animals)

    @property
    def certainty_P2(self) -> float:
        """Indicating the certainty of the identity following the P2"""

        if self.P2_vector is None or self.P2_vector.sum() < 0.001:
            return 0.0

        P2_vector_ordered = np.sort(self.P2_vector)
        P2_first_max = P2_vector_ordered[-1]
        P2_second_max = P2_vector_ordered[-2]

        with np.errstate(divide="ignore"):
            return P2_first_max / P2_second_max

    def set_P1_from_frequencies(self, frequencies: np.ndarray):
        """Given the frequencies of a individual fragment
        computer the P1 vector.

        P1 is the softmax of the frequencies with base 2 for each identity.
        Numpy array indicating the number of images assigned with each of
        the possible identities
        """
        with np.errstate(over="ignore"):
            self.P1_vector = 1.0 / (
                2.0
                ** (
                    np.tile(frequencies, (len(frequencies), 1)).T
                    - np.tile(frequencies, (len(frequencies), 1))
                )
            ).sum(axis=0)

    @staticmethod
    def compute_median_softmax(softmax_probs, number_of_animals):
        """Given the softmax of the predictions outputted by the identification
        network, it computes their median according to the argmax of the
        softmaxed predictions per image.

        Parameters
        ----------
        softmax_probs : ndarray
            array of shape [number_of_images_in_fragment, number_of_animals]
            whose rows are the result of applying the softmax function to the
            predictions outputted by the idCNN per image
        number_of_animals : int
            number of animals to be tracked as defined by the user

        Returns
        -------
        float
            Median of argmax(softmax_probs) per identity

        """
        softmax_probs = np.asarray(softmax_probs)
        # jumps are fragment composed by a single image, thus:
        if len(softmax_probs.shape) == 1:
            softmax_probs = np.expand_dims(softmax_probs, axis=1)
        max_softmax_probs = np.max(softmax_probs, axis=1)
        argmax_softmax_probs = np.argmax(softmax_probs, axis=1)
        softmax_median = np.zeros(number_of_animals)
        for i in np.unique(argmax_softmax_probs):
            softmax_median[i] = np.median(max_softmax_probs[argmax_softmax_probs == i])
        return softmax_median

    def set_certainty_of_individual_fragment(self, median_softmax: np.ndarray):
        """Computes the certainty given the P1_vector of the fragment by
        using the output of :meth:`compute_median_softmax`

        Parameters
        ----------
        P1_vector : numpy array
            Array with shape [1, number_of_animals] computed from frequencies
            by :meth:`compute_identification_statistics`
        median_softmax : ndarray
            Median of argmax(softmax_probs) per image

        Returns
        -------
        float
            Fragment's certainty

        """
        argsort_p1_vector = self.P1_vector.argsort()
        sorted_p1_vector = self.P1_vector[argsort_p1_vector]
        sorted_softmax_probs = median_softmax[argsort_p1_vector]
        certainty = (
            np.diff((sorted_p1_vector * sorted_softmax_probs)[-2:])
            / sorted_p1_vector[-2:].sum()
        )
        self.certainty = certainty[0]

    def get_neighbour_fragment(
        self,
        fragments: list["Fragment"],
        scope: str,
        number_of_frames_in_direction: int = 0,
    ) -> "Fragment | None":
        """If it exist, gets the fragment in the list of all fragment whose
        identity is the identity assigned to self and whose starting frame is
        the ending frame of self + 1, or ending frame is the starting frame of
        self - 1

        Parameters
        ----------
        fragments : list
            List of all the fragments in the video
        scope : str
            If "to_the_future" looks for the consecutive fragment wrt to self,
            if "to_the_past" looks for the fragment the precedes self
        number_of_frames_in_direction : int
            Distance (in frame) at which the previous or next fragment has to
            be

        Returns
        -------
        :class:`fragment.Fragment`
            The neighbouring fragment with respect to self in the direction
            specified by scope if it exists. Otherwise None

        """
        # TODO optimize
        if scope == "to_the_past":
            neighbour = [
                fragment
                for fragment in fragments
                if fragment.is_an_individual
                and len(fragment.assigned_identities) == 1
                and fragment.assigned_identities[0] == self.assigned_identities[0]
                and self.start_frame - fragment.end_frame
                == number_of_frames_in_direction
            ]
        elif scope == "to_the_future":
            neighbour = [
                fragment
                for fragment in fragments
                if fragment.is_an_individual
                and len(fragment.assigned_identities) == 1
                and fragment.assigned_identities[0] == self.assigned_identities[0]
                and fragment.start_frame - self.end_frame
                == number_of_frames_in_direction
            ]
        else:
            raise ValueError(scope)

        assert len(neighbour) < 2
        return neighbour[0] if len(neighbour) == 1 else None

    def set_partially_or_globally_accumulated(self, accumulation_strategy):
        """Sets :attr:`accumulated_globally` and :attr:`accumulated_partially`
        according to `accumulation_strategy`.

        Parameters
        ----------
        accumulation_strategy : str
            Can be "global" or "partial"

        """
        if accumulation_strategy == "global":
            self.accumulated_globally = True
        elif accumulation_strategy == "partial":
            self.accumulated_partially = True

    @property
    def properties(self) -> Sequence[str]:
        return (
            f"Fragment {self.identifier}",
            (
                f"Frames from {self.start_frame} to {self.end_frame} (length"
                f" {self.end_frame-self.start_frame})"
            ),
            ("Individual" if self.is_an_individual else "Crossing")
            + " fragment"
            + (" (forced)" if self.forced_crossing else ""),
            ("Used" if self.used_for_training else "Not used") + " for training",
            ("Used" if self.used_for_pretraining else "Not used") + " for pretraining",
            ("Acceptable" if self.acceptable_for_training else "Not acceptable")
            + " for training",
            f"Predicted identity: {self.identity}",
            f"Corrected solving jumps: {self.identity_corrected_solving_jumps}",
            f"Corrected solving gaps: {self.identities_corrected_closing_gaps}",
            f"Fixed identity: {self.identity_is_fixed}",
            f"Globally accumulated: {self.accumulated_globally}",
            f"Partially accumulated: {self.accumulated_partially}",
            f"Accumulable: {self.accumulable}",
            f"Accumulated at step {self.accumulation_step}",
            f"Is certain: {self.is_certain}",
            "Non consistent" if self.non_consistent else "Consistent",
            (
                f"Max P1 {np.argmax(self.P1_vector)+1} with value"
                f" {self.P1_vector.max()}"
                if hasattr(self, "P1_vector")
                else "Doesn't have P1 vector"
            ),
            f"Certainty: {self.certainty}",
            f"P1 below random: {self.P1_below_random}",
        )



import logging

import numpy as np
from torch.utils.data import DataLoader
from torchvision import transforms
from torchvision.datasets.folder import VisionDataset

from idtrackerai.network import normalize
from idtrackerai.utils import conf

num_workers_train = 1
num_workers_val = 1


class IdentificationDataset(VisionDataset):
    def __init__(self, data_dict, scope, transform=None):
        super().__init__("", transform=transform)
        self.scope = scope
        self.images = data_dict["images"]
        if self.scope in ("training", "validation", "test"):
            self.labels = data_dict["labels"]
        else:
            self.labels = np.zeros((self.images.shape[0]))
        self.get_data()

    def get_data(self):
        if self.images.ndim <= 3:
            self.images = np.expand_dims(np.asarray(self.images), axis=-1)

        if self.scope == "training":
            self.images, self.labels = duplicate_PCA_images(self.images, self.labels)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        image = self.images[index]
        target = self.labels[index]
        if self.transform is not None:
            image = self.transform(image)
        return image, target


def split_data_train_and_validation(images, labels, validation_proportion=None):
    """Splits a set of `images` and `labels` into training and validation sets

    Parameters
    ----------
    number_of_animals : int
        Number of classes in the set of images
    images : list
        List of images (arrays of shape [height, width])
    labels : list
        List of integers from 0 to `number_of_animals` - 1
    validation_proportion : float
        The proportion of images that will be used to create the validation set.


    Returns
    -------
    training_dataset : <DataSet object>
        Object containing the images and labels for training
    validation_dataset : <DataSet object>
        Object containing the images and labels for validation

    See Also
    --------
    :class:`get_data.DataSet`
    :func:`get_data.duplicate_PCA_images`
    """

    if validation_proportion is None:
        validation_proportion = conf.VALIDATION_PROPORTION

    # Init variables
    train_images = []
    train_labels = []
    validation_images = []
    validation_labels = []

    for i in np.unique(labels):
        # Get images of this individual
        this_indiv_images = images[labels == i]
        this_indiv_labels = labels[labels == i]
        # Compute number of images for training and validation
        num_images = len(this_indiv_labels)
        num_images_validation = np.ceil(validation_proportion * num_images).astype(int)
        num_images_training = num_images - num_images_validation
        # Get train, validation and test, images and labels
        train_images.append(this_indiv_images[:num_images_training])
        train_labels.append(this_indiv_labels[:num_images_training])
        validation_images.append(this_indiv_images[num_images_training:])
        validation_labels.append(this_indiv_labels[num_images_training:])

    train_images = np.vstack(train_images)
    train_labels = np.concatenate(train_labels, axis=0)

    validation_images = np.vstack(validation_images)
    validation_labels = np.concatenate(validation_labels, axis=0)

    training_weights = (
        1.0 - np.unique(train_labels, return_counts=True)[1] / len(train_labels)
    ).astype("float32")

    train_dict = {
        "images": train_images,
        "labels": train_labels,
        "weights": training_weights,
    }
    val_dict = {"images": validation_images, "labels": validation_labels}
    return train_dict, val_dict


def duplicate_PCA_images(training_images, training_labels):
    """Creates a copy of every image in `training_images` by rotating 180 degrees

    Parameters
    ----------
    training_images : ndarray
        Array of shape [number of images, height, width, channels] containing
        the images to be rotated
    training_labels : ndarray
        Array of shape [number of images, 1] containing the labels corresponding
        to the `training_images`

    Returns
    -------
    training_images : ndarray
        Array of shape [2*number of images, height, width, channels] containing
        the original images and the images rotated
    training_labels : ndarray
        Array of shape [2*number of images, 1] containing the labels corresponding
        to the original images and the images rotated
    """
    augmented_images = np.rot90(training_images, 2, axes=(1, 2))
    training_images = np.concatenate([training_images, augmented_images], axis=0)
    training_labels = np.concatenate([training_labels, training_labels], axis=0)
    return training_images, training_labels


def get_training_data_loaders(
    number_of_animals: int, train_data, val_data
) -> tuple[DataLoader, DataLoader]:
    logging.info("Creating training IdentificationDataset")
    training_set = IdentificationDataset(
        train_data,
        scope="training",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    train_loader = DataLoader(
        training_set,
        batch_size=conf.BATCH_SIZE_IDCNN,
        shuffle=True,
        num_workers=num_workers_train,
        persistent_workers=num_workers_train > 0,
    )
    train_loader.num_classes = number_of_animals
    train_loader.image_shape = training_set[0][0].shape

    logging.info("Creating validation IdentificationDataset")
    validation_set = IdentificationDataset(
        val_data,
        scope="validation",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    val_loader = DataLoader(
        validation_set,
        batch_size=conf.BATCH_SIZE_PREDICTIONS_IDCNN,
        num_workers=num_workers_val,
        persistent_workers=num_workers_val > 0,
    )
    val_loader.num_classes = number_of_animals
    val_loader.image_shape = validation_set[0][0].shape
    return train_loader, val_loader


def get_test_data_loader(test_data, n_classes):
    logging.debug("Creating test IdentificationDataset")
    test_set = IdentificationDataset(
        test_data,
        scope="predict",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    test_loader = DataLoader(
        test_set,
        batch_size=conf.BATCH_SIZE_PREDICTIONS_IDCNN,
        num_workers=num_workers_val,
        persistent_workers=num_workers_val > 0,
    )
    test_loader.num_classes = n_classes
    test_loader.image_shape = test_set[0][0].shape
    return test_loader



from pathlib import Path
from typing import Sequence

from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtGui import QFocusEvent
from qtpy.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QListView,
    QListWidget,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QWidget,
)

from idtrackerai import Video
from idtrackerai.utils import CustomError
from idtrackerai_GUI_tools import WrappedLabel, key_event_modifier


class AdaptativeList(QListWidget):
    def __init__(self):
        super().__init__()
        self.setAlternatingRowColors(True)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.setMovement(QListView.Movement.Free)
        self.model().rowsInserted.connect(self.set_size)
        self.model().rowsRemoved.connect(self.set_size)

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.set_size()

    def set_size(self):
        content_height = self.sizeHintForRow(0) * max(1, min(5, self.count()))
        scroll_bar = self.horizontalScrollBar()
        self.setFixedHeight(
            content_height
            + 2 * self.frameWidth()
            + (scroll_bar.height() if scroll_bar.isVisible() else 0)
        )

    def keyPressEvent(self, e):
        event = key_event_modifier(e)
        if event is not None:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, e):
        event = key_event_modifier(e)
        if event is not None:
            super().keyReleaseEvent(event)

    def focusOutEvent(self, event: QFocusEvent):
        super().focusOutEvent(event)
        if event.reason() is not Qt.FocusReason.ActiveWindowFocusReason:
            self.clearSelection()  # this IF fixes drag and drop errors on Wayland


class OpenVideoWidget(QWidget):
    new_video_paths = Signal(list, tuple, int, int, list)
    path_clicked = Signal(int)
    video_paths_reordered = Signal(list)
    new_episodes = Signal(list, object)

    def __init__(self, parent, frames_per_episode: int):
        super().__init__()
        self.setLayout(QHBoxLayout())
        self.parent_widget = parent
        self.frames_per_episode = frames_per_episode
        self.button_open = QPushButton("Open video(s)")
        self.button_open.setShortcut("Ctrl+O")
        self.button_open.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.button_open.clicked.connect(self.button_open_clicked)
        self.button_open.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed
        )
        self.list_of_files = AdaptativeList()
        self.list_of_files.model().rowsMoved.connect(self.video_paths_reordered_func)
        self.single_file_label = WrappedLabel(framed=True)
        self.layout().addWidget(self.button_open)
        self.layout().addWidget(self.list_of_files)
        self.layout().addWidget(self.single_file_label)
        self.list_of_files.setVisible(False)
        self.list_of_files.itemSelectionChanged.connect(self.video_path_clicked)
        self.list_of_files.itemClicked.connect(self.video_path_clicked)
        self.single_file_label.setVisible(False)
        self.tracking_intervals = None

    def video_path_clicked(self):
        items = self.list_of_files.selectedItems()
        if items:
            self.path_clicked.emit(self.video_path_start[items[0].text()][0])

    def video_paths_reordered_func(self):
        self.video_path_start.clear()
        i = 0
        for video_path in self.video_paths:
            n_frames = self.video_path_n_frames[video_path]
            self.video_path_start[video_path] = (i, i + n_frames)
            i += n_frames
        self.video_paths_reordered.emit(self.video_paths)
        self.n_frames, video_paths_n_frames, _, self.episodes = (
            Video.get_processing_episodes(
                self.video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )
        self.new_episodes.emit(self.video_paths, self.episodes)

    def button_open_clicked(self):
        video_paths, _ = QFileDialog.getOpenFileNames(
            self.parent_widget, "Open a video file to track"
        )
        self.open_video_paths(sorted(video_paths))

    def open_video_paths(self, video_paths: Sequence[str | Path] | None):
        if not video_paths:
            return
        try:
            video_paths = [Path(path).expanduser().resolve() for path in video_paths]
            Video.assert_video_paths(video_paths)
            self.video_width, self.video_height, self.fps = (
                Video.get_info_from_video_paths(video_paths)
            )
        except (ValueError, CustomError) as e:
            QMessageBox.warning(self, "Video paths error", str(e))
            return

        self.single_file = len(video_paths) == 1
        if self.single_file:
            self.single_file_label.setText(str(video_paths[0]))
        else:
            self.list_of_files.clear()
            self.list_of_files.addItems(map(str, video_paths))

        self.single_file_label.setVisible(self.single_file)
        self.list_of_files.setVisible(not self.single_file)

        self.n_frames, video_paths_n_frames, _, self.episodes = (
            Video.get_processing_episodes(
                video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )
        self.video_path_n_frames = dict(zip(self.video_paths, video_paths_n_frames))

        self.video_path_start = {}
        i = 0
        for video_path in self.video_paths:
            n_frames = self.video_path_n_frames[video_path]
            self.video_path_start[video_path] = (i, i + n_frames)
            i += n_frames

        self.n_frames = i
        self.new_video_paths.emit(
            self.video_paths,
            (self.video_width, self.video_height),
            self.n_frames,
            self.fps,
            self.episodes,
        )

    def set_tracking_interval(self, tracking_intervals):
        self.tracking_intervals = tracking_intervals

        if not hasattr(self, "video_paths"):
            return

        self.n_frames, video_paths_n_frames, _, self.episodes = (
            Video.get_processing_episodes(
                self.video_paths, self.frames_per_episode, self.tracking_intervals
            )
        )
        self.new_episodes.emit(self.video_paths, self.episodes)

    @property
    def video_paths(self):
        return self.getVideoPaths()

    def getVideoPaths(self) -> list[str]:
        if self.single_file:
            return [self.single_file_label.text()]
        return [
            self.list_of_files.item(i).text() for i in range(self.list_of_files.count())
        ]



import logging
import os
import sys
from datetime import datetime
from functools import wraps
from importlib import metadata
from pathlib import Path
from platform import platform
from typing import Callable

from rich.console import Console
from rich.logging import RichHandler

from .check_PyPI_version import check_version_on_console_thread

LOG_FILE_PATH = Path("idtrackerai.log").resolve()


class CustomError(Exception):
    pass


def initLogger(level: int = logging.DEBUG):
    logger_width_when_no_terminal = 130
    try:
        os.get_terminal_size()
    except OSError:
        # stdout is sent to file. We define logger width to a constant
        size = logger_width_when_no_terminal
    else:
        # stdout is sent to terminal
        # We define logger width to adapt to the terminal width
        size = None

    LOG_FILE_PATH.unlink(True)  # avoid conflicts and merged files

    # The first handler is the terminal, the second one the .log file,
    # both rendered with Rich and full logging (level=0)
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="%H:%M:%S",
        force=True,
        handlers=[
            RichHandler(console=Console(width=size)),
            RichHandler(
                console=Console(
                    file=LOG_FILE_PATH.open("w", encoding="utf_8"),  # noqa SIM115
                    width=logger_width_when_no_terminal,
                )
            ),
        ],
    )

    logging.captureWarnings(True)
    logging.info("Welcome to idtracker.ai")
    logging.debug(
        f"Running idtracker.ai '{metadata.version('idtrackerai')}'"
        f" on Python '{sys.version.split(' ')[0]}'\nPlatform: '{platform(True)}'"
        "\nDate: "
        + str(datetime.now()).split(".")[0]
    )
    logging.info("Writing log in %s", LOG_FILE_PATH)


def wrap_entrypoint(main_function: Callable):
    @wraps(main_function)
    def ret_fun(*args, **kwargs):
        initLogger()
        check_version_on_console_thread()
        try:
            return main_function(*args, **kwargs)
        except CustomError as error:
            logging.critical(error, exc_info=False)
            return False
        except Exception as error:
            logging.critical("%s: %s", type(error).__name__, error, exc_info=True)
            logging.info(
                "\n\nIf this error persists please let us know by "
                "following any of the following options:\n"
                "  - posting on "
                "https://groups.google.com/g/idtrackerai_users\n"
                "  - opening an issue at "
                "https://gitlab.com/polavieja_lab/idtrackerai\n"
                "  - sending an email to idtrackerai@gmail.com\n"
                "Share the log file (%s) when "
                "doing any of the options above",
                LOG_FILE_PATH,
            )
            return False

    return ret_fun



import sys
from argparse import ArgumentParser
from pathlib import Path

from qtpy.QtWidgets import QApplication

from idtrackerai.utils import wrap_entrypoint
from idtrackerai_validator.validation_GUI import ValidationGUI


def input_args():
    parser = ArgumentParser()
    parser.add_argument(
        "session_directory", help="Session directory to validate", type=Path, nargs="?"
    )
    return parser.parse_args()


@wrap_entrypoint
def main():
    args = input_args()
    app = QApplication(sys.argv)
    window = ValidationGUI(args.session_directory)
    window.show()
    app.exec()



import logging
import sys
from contextlib import suppress
from pathlib import Path

import numpy as np
import torch
from rich.console import Console
from rich.status import Status
from torch.utils.data import DataLoader

from idtrackerai import Blob
from idtrackerai.network import (
    DEVICE,
    LearnerClassification,
    NetworkParams,
    evaluate,
    train,
)
from idtrackerai.utils import conf, track

from .crossings_dataset import get_test_data_loader


class StopTraining:
    """CROSSING Stops the training of the network according to the conditions specified
    in __call__
    """

    epochs_before_checking_stopping_conditions = 10

    def __init__(self, num_epochs: int):
        logging.info("Setting the stopping criteria", stacklevel=3)
        self.num_epochs = num_epochs  # maximal num of epochs
        self.overfitting_counter: int = 0
        """Number of epochs in which the network is overfitting before
        stopping the training"""

        self.epochs_completed = -1

    def __call__(
        self,
        loss_training: float,
        loss_validation: list,
        accuracy_validation: float,
        status: Status,
    ):
        """Returns True when one of the conditions to stop the training is
        satisfied, otherwise it returns False"""
        self.epochs_completed += 1

        if self.epochs_completed > 0 and (
            np.isnan(loss_training) or np.isnan(loss_validation[-1])
        ):
            status.stop()
            logging.error(
                "The model diverged with loss NaN, falling back "
                "to detecting crossings with the model area"
            )
            return True
        # check if it did not reached the epochs limit
        if self.epochs_completed >= self.num_epochs:
            status.stop()
            logging.warning(
                "The number of epochs completed is larger than the number "
                "of epochs set for training, we stop the training"
            )
            return True

        if self.epochs_completed <= self.epochs_before_checking_stopping_conditions:
            return False

        # check that the model is not overfitting or if it reached
        # a stable saddle (minimum)
        current_loss = loss_validation[-1]
        previous_loss = np.nanmean(
            loss_validation[-self.epochs_before_checking_stopping_conditions : -1]
        )

        # The validation loss in the first 10 epochs could have exploded
        # but being decreasing.
        if np.isnan(previous_loss):
            previous_loss = sys.float_info[0]
        losses_difference = previous_loss - current_loss

        # check overfitting
        if losses_difference < 0.0:
            self.overfitting_counter += 1
            if self.overfitting_counter >= conf.OVERFITTING_COUNTER_THRESHOLD_DCD:
                status.stop()
                logging.info("Overfitting")
                return True
        else:
            self.overfitting_counter = 0
        # check if the error is not decreasing much
        if np.abs(losses_difference) < conf.LEARNING_PERCENTAGE_DIFFERENCE_DCD * 10 ** (
            int(np.log10(current_loss)) - 1
        ):
            status.stop()
            logging.info("The losses difference is very small, we stop the training")
            return True
        # if the individual accuracies in validation are 1. for all the animals
        if accuracy_validation == 1.0:
            status.stop()
            logging.info("The accuracy in validation is 100%, we stop the training")
            return True
        # if the validation loss is 0.
        if previous_loss == 0.0 or current_loss == 0.0:
            status.stop()
            logging.info("The validation loss is 0.0, we stop the training")
            return True

        return False


def train_deep_crossing(
    learner: LearnerClassification,
    train_loader: DataLoader,
    val_loader: DataLoader,
    network_params: NetworkParams,
    stop_training: StopTraining,
) -> tuple[bool, Path]:
    logging.info("Training Deep Crossing Detector")

    # Initialize metric storage
    train_loss = 0.0
    val_loss = 0.0
    val_losses = []
    val_acc = 0.0

    logging.debug("Entering the epochs loop...")
    with Console().status("[red]Epochs loop...") as status:
        while not stop_training(train_loss, val_losses, val_acc, status):
            epoch = stop_training.epochs_completed

            train_loss = train(epoch, train_loader, learner)
            val_loss, val_acc = evaluate(val_loader, network_params.n_classes, learner)

            val_losses.append(val_loss)

            with suppress(IndexError):
                status.update(
                    f"[red]Epoch {epoch}: training loss ="
                    f" {train_loss:.6f}, validation loss ="
                    f" {val_loss:.6f} and accuracy = {val_acc:.4%}"
                )

        logging.info("Last epoch loop: %s", status.status, extra={"markup": True})

    learner.save_model(network_params.model_path)
    return np.isnan(train_loss) or np.isnan(val_loss), network_params.model_path


def get_predictions_crossigns(
    id_images_file_paths: list[Path], model: torch.nn.Module, blobs: list[Blob]
):
    loader = get_test_data_loader(id_images_file_paths, blobs)
    predictions = []

    model.eval()
    with torch.no_grad():
        for input, _target in track(loader, "Predicting crossings"):
            # Prepare the inputs

            # Inference
            output = model(input.to(DEVICE))
            pred = output.argmax(1)  # find the predicted class

            predictions += pred.tolist()

    return predictions



from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from idtrackerai_GUI_tools import WrappedLabel

Selected_Color = QColor(255, 0, 0)
Unselected_Color = QColor(255, 255, 255)
Selected_Color_alpha = QColor(255, 0, 0, 75)
Unselected_Color_alpha = QColor(255, 255, 255, 75)


class IdGroups(QWidget):
    needToDraw = Signal()
    unsaved_changes = Signal()

    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)
        first_row = QHBoxLayout()
        first_row.addWidget(QLabel("Identity groups"))
        self.add_btn = QToolButton()
        self.add_btn.setText("Add")
        first_row.addWidget(self.add_btn)
        self.main_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.main_layout.addLayout(first_row)
        self.add_btn.clicked.connect(self.add_clicked)
        self.editing_name: str = ""
        self.view: set[str] = set()
        self.id_groups: dict[str, tuple[QWidget, set[int]]] = {}

    def generate_row(self, name: str, group: set[int]):
        label = WrappedLabel(f"{name}: {', '.join(map(str,group))}")
        label.setObjectName("label")
        label.setWordWrap(True)
        view_btn = QToolButton()
        view_btn.setText("View")
        view_btn.setObjectName("view")
        view_btn.setCheckable(True)
        view_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        view_btn.toggled.connect(
            lambda c: self.view_btn_clicked(view_btn, label.text().split(":")[0], c)
        )
        edit_btn = QToolButton()
        edit_btn.setText("Edit")
        edit_btn.setObjectName("edit")
        edit_btn.setCheckable(True)
        edit_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        edit_btn.toggled.connect(
            lambda c: self.edit_btn_clicked(edit_btn, label.text().split(":")[0], c)
        )
        remove_btn = QToolButton()
        remove_btn.setText("-")
        remove_btn.setObjectName("remove")
        remove_btn.setToolTip("Remove")
        remove_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        remove_btn.clicked.connect(
            lambda: self.remove_btn_clicked(label.text().split(":")[0])
        )

        row = QWidget()
        layout = QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(view_btn)
        layout.addWidget(edit_btn)
        layout.addWidget(remove_btn)
        row.setLayout(layout)
        self.main_layout.addWidget(row)
        return row

    def uncheck_btns(self, exception: QToolButton | None):
        for btns, _group in self.id_groups.values():
            for widget in btns.findChildren(QToolButton):
                assert isinstance(widget, QToolButton)
                if widget != exception and widget.isChecked():
                    widget.setChecked(False)

    def view_btn_clicked(self, btn: QToolButton, name: str, checked: bool):
        if checked:
            self.uncheck_btns(btn)
            self.view.add(name)
        else:
            self.view.remove(name)
        self.needToDraw.emit()

    def edit_btn_clicked(self, btn: QToolButton, name: str, checked: bool):
        if checked:
            self.uncheck_btns(btn)
        self.editing_name = name if checked else ""
        self.needToDraw.emit()

    def remove_btn_clicked(self, name: str):
        self.uncheck_btns(None)
        row = self.id_groups.pop(name)[0]
        self.main_layout.removeWidget(row)

    def load_groups(self, identities_groups: dict):
        while self.id_groups:
            row = self.id_groups.popitem()[1][0]
            self.main_layout.removeWidget(row)

        self.id_groups = {
            key: (self.generate_row(key, set(value)), set(value))
            for key, value in identities_groups.items()
        }

    def selected_id(self, identity: int | None):
        if not self.editing_name or identity in (-1, None):
            return

        row, group = self.id_groups[self.editing_name]
        if identity in group:
            group.remove(identity)
        else:
            group.add(identity)
        label = row.findChild(WrappedLabel, "label")
        assert isinstance(label, WrappedLabel)
        label.setText(f"{self.editing_name}: {', '.join(map(str,group))}")
        self.unsaved_changes.emit()

    def add_clicked(self):
        name, ok = QInputDialog.getText(
            self, "idtracker.ai", "Enter identity group name:"
        )
        name = name.strip()

        if not ok or not name:
            return

        if name in self.id_groups:
            edit_btn = self.id_groups[name][0].findChild(QToolButton, "edit")
            assert isinstance(edit_btn, QToolButton)
            edit_btn.setChecked(True)
            return

        btns, group = self.generate_row(name, set()), set()
        self.id_groups[name] = btns, group
        edit = btns.findChild(QToolButton, "edit")
        assert isinstance(edit, QToolButton)
        edit.setChecked(True)

    def is_active(self) -> bool:
        return (
            self.isVisible()
            and self.isEnabled()
            and (bool(self.editing_name) or bool(self.view))
        )

    def get_groups(self):
        return {key: value[1] for key, value in self.id_groups.items()}

    def get_cmaps(self, n_animals: int):
        names = [self.editing_name] if self.editing_name else self.view
        cmap = [Unselected_Color] * (n_animals + 1)
        cmap_alpha = [Unselected_Color_alpha] * (n_animals + 1)

        for name in names:
            group = self.id_groups[name][1]
            for identity in group:
                cmap[identity] = Selected_Color
                cmap_alpha[identity] = Selected_Color_alpha

        return cmap, cmap_alpha

    def uncheck_edit_buttons(self):
        for edit_btn in self.findChildren(QToolButton, "edit"):
            assert isinstance(edit_btn, QToolButton)
            edit_btn.setChecked(False)



import numpy as np
from qtpy.QtCore import Qt, QThread, QTimer, Signal  # type: ignore
from qtpy.QtGui import QImage, QPainter, QPixmap
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QHBoxLayout,
    QMessageBox,
    QProgressDialog,
    QToolButton,
    QWidget,
)

from idtrackerai.animals_detection.segmentation import (
    generate_background_from_frame_stack,
    generate_frame_stack,
)
from idtrackerai_GUI_tools import Canvas


class BkgComputationThread(QThread):
    progress_changed = Signal(int)

    def __init__(self, n_frames_for_background: int, background_stat: str):
        super().__init__()
        self.frame_stack = None
        self.bkg = None
        self.abort = False
        self.n_frames_for_background = n_frames_for_background
        self.background_stat = background_stat
        self.finished.connect(
            lambda: self.progress_changed.emit(n_frames_for_background)
        )

    def set_parameters(self, video_paths, episodes):
        self.video_paths = video_paths
        self.episodes = episodes

    def run(self):
        self.abort = False
        if self.bkg is None:
            if self.frame_stack is None:
                self.frame_stack = generate_frame_stack(
                    self.video_paths,
                    self.episodes,
                    self.n_frames_for_background,
                    self.progress_changed,
                    lambda: self.abort,
                )
            if self.abort:
                self.frame_stack = None
                self.abort = False
                return

            self.bkg = generate_background_from_frame_stack(
                self.frame_stack, self.background_stat
            )

            if self.abort:
                self.frame_stack = None
                self.bkg = None
                self.abort = False
                return

    def quit(self):
        self.abort = True


class ImageDisplay(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Background")
        self.canvas = Canvas()
        self.canvas.painting_time.connect(self.paint_image)

        self.setLayout(QHBoxLayout())
        self.setMinimumSize(200, 50)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.canvas)

    def paint_image(self, painter: QPainter):
        painter.drawPixmap(0, 0, self.pixmap)

    def show(self, frame: np.ndarray):
        height, width = frame.shape
        self.pixmap = QPixmap.fromImage(
            QImage(frame.data, width, height, width, QImage.Format.Format_Grayscale8)
        )

        self.canvas.centerX = int(width / 2)
        self.canvas.centerY = int(height / 2)

        ratio = width / height

        QDialog_size = 600
        if width > height:
            window_width = QDialog_size
            window_height = int(QDialog_size / ratio)
        else:
            window_width = int(QDialog_size / ratio)
            window_height = QDialog_size
        self.setGeometry(0, 0, window_width, window_height)
        QTimer.singleShot(0, lambda: self.canvas.adjust_zoom_to(width, height))
        super().exec()


class BkgWidget(QWidget):
    new_bkg_data = Signal(object)

    def __init__(
        self, parent: QWidget, n_frames_for_background: int, background_stat: str
    ):
        super().__init__()
        self.checkBox = QCheckBox("Background subtraction")
        self.checkBox.stateChanged.connect(self.CheckBox_changed)
        self.view_bkg = QToolButton()
        self.view_bkg.setText("View background")
        self.bkg_thread = BkgComputationThread(n_frames_for_background, background_stat)
        self.view_bkg.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.view_bkg.setVisible(False)
        self.progress_bar = QProgressDialog(
            "Computing background", "Cancel", 0, n_frames_for_background, parent
        )
        self.progress_bar.cancel()
        self.progress_bar.setMinimumDuration(1000)
        self.progress_bar.setModal(True)
        self.progress_bar.canceled.connect(self.bkg_thread.quit)
        self.view_bkg.clicked.connect(self.view_bkg_clicked)

        self.image_display = ImageDisplay(parent)
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.checkBox)
        layout.addWidget(self.view_bkg)
        self.bkg_thread.progress_changed.connect(self.progress_bar.setValue)
        self.bkg_thread.finished.connect(self.bkg_thread_finished)

    def set_new_video_paths(self, video_paths, episodes):
        self.video_paths = video_paths
        self.episodes = episodes
        self.bkg_thread.bkg = None
        self.bkg_thread.frame_stack = None
        if not self.checkBox.isChecked():
            return
        QMessageBox.information(
            self,
            "Background deactivated",
            "The subtracted background depends on the specified video paths. Check"
            " again the background subtraction if desired when finish editing the"
            " video paths.",
        )
        self.checkBox.setChecked(False)

    def view_bkg_clicked(self):
        if self.bkg_thread.bkg is not None:
            self.image_display.show(self.bkg_thread.bkg)

    def CheckBox_changed(self, checked):
        if checked:
            if not hasattr(self, "video_paths"):
                self.checkBox.setChecked(False)
                return
            self.bkg_thread.set_parameters(self.video_paths, self.episodes)
            self.bkg_thread.start()
        else:
            self.view_bkg.setVisible(False)
            self.new_bkg_data.emit(None)

    def bkg_thread_finished(self):
        if self.bkg_thread.bkg is None:
            self.checkBox.setChecked(False)
            self.view_bkg.setVisible(False)
        else:
            self.view_bkg.setVisible(True)
        self.new_bkg_data.emit(self.bkg_thread.bkg)

    def getBkg(self):
        return self.bkg_thread.bkg if self.checkBox.isChecked() else None



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
from typing import Callable

import numpy as np

from idtrackerai import Blob, Fragment, Video
from idtrackerai.utils import track


def produce_trajectories(
    blobs_in_video: list[list[Blob]],
    number_of_animals: int,
    progress_bar=None,
    abort: Callable = lambda: False,
    fragments: list[Fragment] | None = None,
):
    """Produce trajectories array from ListOfBlobs

    Parameters
    ----------
    blobs_in_video : <ListOfBlobs object>
        See :class:`list_of_blobs.ListOfBlobs`
    number_of_frames : int
        Total number of frames in video
    number_of_animals : int
        Number of animals to be tracked

    Returns
    -------
    dict
        Dictionary with np.array as values (trajectories organized by identity)

    """
    number_of_frames = len(blobs_in_video)
    centroid_trajectories = np.full((number_of_frames, number_of_animals, 2), np.NaN)
    id_probabilities = np.full((number_of_frames, number_of_animals, 1), np.NaN)
    areas = np.full((number_of_frames, number_of_animals), np.NaN)

    for frame_number, blobs_in_frame in enumerate(
        track(blobs_in_video, "Producing trajectories")
    ):
        if abort():
            return None, None, {}
        if progress_bar:
            progress_bar.emit(frame_number)
        for blob in blobs_in_frame:
            for identity, centroid in blob.final_ids_and_centroids:
                if identity not in (None, 0):
                    centroid_trajectories[blob.frame_number, identity - 1, :] = centroid
            blob_final_identities = list(blob.final_identities)
            if blob.is_an_individual and len(blob_final_identities) == 1:
                identity = blob_final_identities[0]
                if identity in (None, 0):
                    continue

                areas[blob.frame_number, identity - 1] = blob.area

                if fragments is None:
                    continue
                P2_vector = fragments[blob.fragment_identifier].P2_vector

                if P2_vector is None:
                    continue
                id_probabilities[blob.frame_number, identity - 1, :] = np.max(P2_vector)

    return (
        centroid_trajectories,
        id_probabilities,
        {
            "mean": np.nanmean(areas, axis=0),
            "median": np.nanmedian(areas, axis=0),
            "std": np.nanstd(areas, axis=0),
        },
    )


def produce_trajectories_wo_identification(
    blobs_in_video: list[list[Blob]],
    number_of_animals: int,
    progress_bar=None,
    abort: Callable = lambda: False,
):
    number_of_frames = len(blobs_in_video)
    centroid_trajectories = np.full((number_of_frames, number_of_animals, 2), np.nan)
    identifiers_prev = np.full(number_of_animals, np.nan)
    areas = np.full((number_of_frames, number_of_animals), np.nan)

    for frame_number, blobs_in_frame in enumerate(
        track(blobs_in_video, "Creating trajectories")
    ):
        if abort():
            return None, None, {}
        if progress_bar:
            progress_bar.emit(frame_number)
        try:
            identifiers_next = {
                b.fragment_identifier for b in blobs_in_video[frame_number + 1]
            }
        except IndexError:  # last frame
            identifiers_next = {b.fragment_identifier for b in blobs_in_frame}

        for blob in blobs_in_frame:
            if blob.is_an_individual:
                if blob.fragment_identifier in identifiers_prev:
                    column = np.argwhere(identifiers_prev == blob.fragment_identifier)[
                        0
                    ][0]
                else:
                    column = np.argwhere(np.isnan(identifiers_prev))[0][0]
                    identifiers_prev[column] = blob.fragment_identifier

                blob.identity = column + 1
                # blobs that are individual only have one centroid
                centroid_trajectories[frame_number, column, :] = next(
                    blob.final_centroids
                )
                areas[frame_number, column] = blob.area

                if blob.fragment_identifier not in identifiers_next:
                    identifiers_prev[column] = np.nan
    return (
        centroid_trajectories,
        None,
        {
            "mean": np.nanmean(areas, axis=0),
            "median": np.nanmedian(areas, axis=0),
            "std": np.nanstd(areas, axis=0),
        },
    )


def produce_output_dict(
    blobs_in_video: list[list[Blob]],
    video: Video,
    fragments: list[Fragment] | None = None,
    progress_bar=None,
    abort: Callable = lambda: False,
):
    """Outputs the dictionary with keys: trajectories, git_commit, video_path,
    frames_per_second

    Parameters
    ----------
    blobs_in_video : list
        List of all blob objects (see :class:`~blob.Blobs`) generated by
        considering all the blobs segmented from the video
    video : <Video object>
        See :class:`~video.Video`

    Returns
    -------
    dict
        Output dictionary containing trajectories as values

    """
    if video.track_wo_identities:
        video.number_of_animals = max(map(len, blobs_in_video))

    centroid_trajectories, id_probabilities, area_stats = (
        produce_trajectories_wo_identification(
            blobs_in_video, video.n_animals, progress_bar, abort
        )
        if video.track_wo_identities
        else produce_trajectories(
            blobs_in_video, video.n_animals, progress_bar, abort, fragments
        )
    )

    if centroid_trajectories is None or abort():
        return None

    output_dict = {
        "trajectories": centroid_trajectories / video.resolution_reduction,
        "version": video.version,
        "video_paths": list(map(str, video.video_paths)),
        "frames_per_second": video.frames_per_second,
        "body_length": video.median_body_length_full_resolution,
        "stats": {"estimated_accuracy": video.estimated_accuracy},
        "areas": area_stats,
        "setup_points": video.setup_points,
        "identities_labels": video.identities_labels or [
            str(i + 1) for i in range(video.n_animals)
        ],
        "identities_groups": {
            key: list(value) for key, value in video.identities_groups.items()
        },
    }

    if id_probabilities is not None and np.isfinite(id_probabilities).any():
        output_dict["id_probabilities"] = id_probabilities
        # After the interpolation some identities that were 0 are assigned
        output_dict["stats"]["estimated_accuracy_after_interpolation"] = (
            1 if video.single_animal else np.nanmean(output_dict["id_probabilities"])
        )
        # Centroids with identity
        identified = ~np.isnan(output_dict["trajectories"][..., 0])
        output_dict["stats"]["percentage_identified"] = np.mean(identified)
        # Estimated accuracy of identified blobs

        output_dict["stats"]["estimated_accuracy_identified"] = (
            1
            if video.single_animal
            else np.nanmean(output_dict["id_probabilities"][identified])
        )

    return output_dict



from statistics import fmean

import torch
from torch.utils.data import DataLoader

from . import DEVICE, LearnerClassification


def train(epoch: int, train_loader: DataLoader, learner: LearnerClassification):
    """Trains trains a network using a learner, a given train_loader"""
    losses = []

    learner.train()

    for input, target in train_loader:
        loss = learner.learn(input.to(DEVICE), target.to(DEVICE))

        losses += [loss] * len(input)

    learner.step_schedule(epoch)
    return fmean(losses)


def evaluate(eval_loader: DataLoader, n_classes: int, learner: LearnerClassification):
    with torch.no_grad():
        # Initialize all meters
        losses = []
        confusion = Confusion(n_classes)

        learner.eval()

        for input, target in eval_loader:
            # Prepare the inputs
            target = target.to(DEVICE)

            # Optimization
            loss, output = learner.forward_with_criterion(input.to(DEVICE), target)

            losses += [loss] * len(input)

            # Update the performance meter
            confusion.add(output, target)

    return fmean(losses), confusion.acc()


class Confusion:
    """
    column of confusion matrix: predicted index
    row of confusion matrix: target index
    """

    def __init__(self, n_classes: int):
        self.k = n_classes
        self.conf = torch.LongTensor(n_classes, n_classes)
        self.conf.fill_(0)

    def add(self, output: torch.Tensor, target: torch.Tensor):
        if target.size(0) > 1:
            output = output.squeeze()
            target = target.squeeze()
        assert output.size(0) == target.size(0)
        if output.ndimension() > 1:  # it is the raw probabilities over classes
            assert output.size(1) == self.conf.size(
                0
            ), "number of outputs does not match size of confusion matrix"

            _, pred = output.max(1)  # find the predicted class
        else:  # it is already the predicted class
            pred = output
        indices = (
            target * self.conf.stride(0) + pred.squeeze_().type_as(target)
        ).type_as(self.conf)
        ones = torch.ones(1).type_as(self.conf).expand(indices.size(0))
        self.conf.view(-1).index_add_(0, indices, ones)

    def acc(self):
        TP = self.conf.diag().sum().item()
        total = self.conf.sum().item()
        return 0.0 if total == 0 else TP / total



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import numpy as np

from . import Blob, Fragment
from .utils import load_id_images


class GlobalFragment:
    """Representes a collection of :class:`fragment.Fragment` N different
    animals. Where N is the number of animals in the video as defined by the
    user.

        Parameters
    ----------
    blobs_in_video : list
        List of lists of instances of :class:`blob.Blob`.
    fragments : list
        List of lists of instances of the class :class:`fragment.Fragment`
    first_frame_of_the_core : int
        First frame of the core of the global fragment. See also
        :func:`list_of_global_fragments.detect_global_fragments_core_first_frame`.
        This also acts as a unique identifier of the global fragment.
    number_of_animals : int
        Number of animals to be tracked as defined by the user.
    """

    accumulation_step: int | None = None
    """Integer indicating the accumulation step at which the fragment was
    accumulated. See also the accumulation_manager.py module."""

    duplicated_identities: set
    first_frame_of_the_core: int
    individual_fragments_identifiers: tuple[int]
    individual_fragments: list[Fragment]
    minimum_distance_travelled: float

    def __init__(
        self,
        blobs_in_video: list[list[Blob]],
        fragments: list[Fragment],
        first_frame_of_the_core: int,
    ):
        self.first_frame_of_the_core = first_frame_of_the_core
        self.individual_fragments_identifiers = tuple(
            blob.fragment_identifier for blob in blobs_in_video[first_frame_of_the_core]
        )
        self.set_individual_fragments(fragments)

        for fragment in self.individual_fragments:
            fragment.is_in_a_global_fragment = True

        self.minimum_distance_travelled = min(
            fragment.distance_travelled for fragment in self.individual_fragments
        )

    @property
    def min_n_images_per_fragment(self):
        return min(fragment.number_of_images for fragment in self.individual_fragments)

    @classmethod
    def from_json(cls, data: dict, fragments: list[Fragment] | None):
        global_fragment = cls.__new__(cls)
        global_fragment.__dict__.update(data)
        if "duplicated_identities" in data:
            global_fragment.duplicated_identities = set(data["duplicated_identities"])

        if fragments is not None:
            global_fragment.set_individual_fragments(fragments)

        return global_fragment

    @property
    def used_for_training(self):
        """Boolean indicating if all the fragments in the global fragment
        have been used for training the identification network"""
        return all(fragment.used_for_training for fragment in self.individual_fragments)

    def is_unique(self, number_of_animals: int):
        """Boolean indicating that the global fragment has unique
        identities, i.e. it does not have duplications."""
        return (
            len(
                set(range(number_of_animals))
                - {fragment.temporary_id for fragment in self.individual_fragments}
            )
            == 0
        )

    @property
    def is_partially_unique(self):
        """Boolean indicating that a subset of the fragments in the global
        fragment have unique identities"""

        identities_acceptable_for_training = [
            fragment.temporary_id
            for fragment in self.individual_fragments
            if fragment.acceptable_for_training
        ]
        self.duplicated_identities = {
            x
            for x in identities_acceptable_for_training
            if identities_acceptable_for_training.count(x) > 1
        }
        return len(self.duplicated_identities) == 0

    def set_individual_fragments(self, fragments: list[Fragment]):
        """Gets the list of instances of the class :class:`fragment.Fragment`
        that constitute the global fragment and sets an attribute with such
        list.

        Parameters
        ----------
        fragments : list
            All the fragments extracted from the video.

        """
        self.individual_fragments = [
            fragments[identifier]
            for identifier in self.individual_fragments_identifiers
        ]

    def acceptable_for_training(self, accumulation_strategy: str) -> bool:
        """Returns True if the global fragment is acceptable for training"""

        return (all if accumulation_strategy == "global" else any)(
            fragment.acceptable_for_training for fragment in self.individual_fragments
        )

    @property
    def total_number_of_images(self) -> int:
        """Gets the total number of images in the global fragment"""
        return sum(fragment.number_of_images for fragment in self.individual_fragments)

    def get_images_and_labels(self, id_images_file_paths):
        """Gets the images and identities in the global fragment as a
        labelled dataset in order to train the identification neural network

        If the scope is "pretraining" the identities of each fragment
        will be arbitrary.
        If the scope is "identity_transfer" then the labels will be
        empty as they will be infered by the identification network selected
        by the user to perform the transferring of identities.

        Parameters
        ----------
        id_images_file_paths : list
            List of paths (str) where the identification images are stored.
        scope : str, optional
            Whether the images are going to be used for training the
            identification network or for "pretraining", by default
            "pretraining".

        Returns
        -------
        Tuple
            Tuple with two Numpy arrays with the images and their labels.
        """
        images = []
        labels = []

        for temporary_id, fragment in enumerate(self.individual_fragments):
            images += fragment.image_locations
            labels += [temporary_id] * fragment.number_of_images

        return load_id_images(id_images_file_paths, images), np.asarray(labels)



"""The correct_impossible_velocity_jumps module"""
# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)

import numpy as np

from idtrackerai import Fragment, ListOfFragments, Video
from idtrackerai.utils import track


def get_candidate_identities_by_minimum_speed(
    fragment: Fragment,
    fragments: list[Fragment],
    available_identities: list[int],
    impossible_velocity_threshold: float,
) -> tuple[np.ndarray, np.ndarray]:
    """Computes the candidate identities for a given `fragment` taking into
    consideration the velocities needed to join the `fragment` with its neighbour
    fragments in the past and in the future

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    available_identities : set
        Set with the available idenities for the `fragment`
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this threshold
        the identiy of one of the fragments is considerd to be wrong as it would be
        physically impossible for an animal to move so much. See `video.velocity_threshold`
        for each definition

    Returns
    -------
    candidate_identities_by_speed : nd.array
        Array with the identities that fullfill the `impossible_velocity_threshold`
        ordered from minimum to maximum velocity

    speed_of_candidate_identities : nd.array
        Array with the maximum velocity needed to link the given `fragment`
        with its neighbours assuming a given identity. Ordered from minimum to maximum
        velocity

    See Also
    --------
    Fragment
    compute_velocities_consecutive_fragments

    """
    speed_of_candidate_identities: list[float] = []
    for identity in available_identities:
        fragment.user_generated_identity = identity
        neighbour_fragment_past = fragment.get_neighbour_fragment(
            fragments, "to_the_past"
        )
        neighbour_fragment_future = fragment.get_neighbour_fragment(
            fragments, "to_the_future"
        )
        velocities_between_fragments = compute_velocities_consecutive_fragments(
            neighbour_fragment_past, fragment, neighbour_fragment_future
        )

        if np.isnan(velocities_between_fragments).all():
            speed_of_candidate_identities.append(impossible_velocity_threshold)
        else:
            speed_of_candidate_identities.append(
                np.nanmax(velocities_between_fragments)
            )
    fragment.user_generated_identity = None
    argsort_identities_by_speed = np.argsort(speed_of_candidate_identities)
    return (
        np.asarray(available_identities)[argsort_identities_by_speed],
        np.asarray(speed_of_candidate_identities)[argsort_identities_by_speed],
    )


def get_candidate_identities_above_random_P2(
    fragment: Fragment,
    fragments: list[Fragment],
    non_available_identities: np.ndarray,
    available_identities: list[int],
    impossible_velocity_threshold: float,
    number_of_animals: int,
):
    """Computes the candidate identities of a `fragment` taking into
    consideration the probability of identification given by its
    `fragment.P2_vector`. An identity is a potential candidate if the
    probability of identification is above random.

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    non_available_identities : nd.array
        Array with the non available identities for the `fragment`
    available_identities : set
        Set with the available idenities for the `fragment`
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this
        threshold the identiy of one of the fragments is considerd to be
        wrong as it would be physically impossible for an animal to move so
        much. See `video.velocity_threshold` for each definition

    Returns
    -------
    candidate_identities_by_speed : nd.array
        Array with the identities that fullfill the
        `impossible_velocity_threshold` ordered from minimum to maximum
        velocity

    speed_of_candidate_identities : nd.array
        Array with the maximum velocity needed to link the given `fragment`
        with its neighbours assuming a given identity. Ordered from
        minimum to maximum velocity

    See Also
    --------
    Fragment
    get_candidate_identities_by_minimum_speed

    """
    P2_vector = fragment.P2_vector
    assert P2_vector is not None
    if len(non_available_identities) > 0:
        P2_vector[non_available_identities - 1] = 0
    if all(P2_vector == 0):
        candidate_identities_speed, _ = get_candidate_identities_by_minimum_speed(
            fragment, fragments, available_identities, impossible_velocity_threshold
        )
        return candidate_identities_speed

    if fragment.number_of_images == 1:
        random_threshold = 1 / number_of_animals
    else:
        random_threshold = 1 / fragment.number_of_images
    return np.argwhere(P2_vector > random_threshold)[:, 0] + 1


def reassign(
    fragment: Fragment,
    list_of_fragments: ListOfFragments,
    impossible_velocity_threshold: float,
):
    """Reassigns the identity of a given `fragment` considering the identity of the
    `fragments` coexisting with it and the `impossible_velocity_threshold`

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this threshold
        the identiy of one of the fragments is considerd to be wrong as it would be
        physically impossible for an animal to move so much. See `video.velocity_threshold`
        for each definition

    See Also
    --------
    :class:`fragment.Fragment`
    :meth:`get_available_and_non_available_identities`
    :meth:`get_candidate_identities_by_minimum_speed`
    :meth:`get_candidate_identities_above_random_P2`

    """

    non_available_identities = {
        coexisting_fragment.assigned_identities[0]
        for coexisting_fragment in fragment.coexisting_individual_fragments
    }
    available_identities = (
        set(range(1, list_of_fragments.n_animals + 1)) - non_available_identities
    )
    if fragment.assigned_identities[0] not in (None, 0):
        available_identities.add(fragment.assigned_identities[0])

    if 0 in non_available_identities:
        non_available_identities.remove(0)
    non_available_identities = np.asarray(list(non_available_identities))

    if len(available_identities) == 1:
        candidate_id = available_identities.pop()
    else:
        available_identities = list(available_identities)
        candidate_identities_speed, speed_of_candidate_identities = (
            get_candidate_identities_by_minimum_speed(
                fragment,
                list_of_fragments.fragments,
                available_identities,
                impossible_velocity_threshold,
            )
        )
        candidate_identities_P2 = get_candidate_identities_above_random_P2(
            fragment,
            list_of_fragments.fragments,
            non_available_identities,
            available_identities,
            impossible_velocity_threshold,
            list_of_fragments.n_animals,
        )
        candidate_identities: list[int] = []
        candidate_speeds: list[float] = []
        for candidate_id, candidate_speed in zip(
            candidate_identities_speed, speed_of_candidate_identities
        ):
            if candidate_id in candidate_identities_P2:
                candidate_identities.append(candidate_id)
                candidate_speeds.append(candidate_speed)
        if not candidate_identities:
            candidate_id = 0
        elif len(candidate_identities) == 1:
            if candidate_speeds[0] < impossible_velocity_threshold:
                candidate_id = candidate_identities[0]
            else:
                candidate_id = 0
        elif len(candidate_identities) > 1:
            if np.count_nonzero(candidate_speeds == np.min(candidate_speeds)) == 1:
                if candidate_speeds[0] < impossible_velocity_threshold:
                    candidate_id = candidate_identities[0]
                else:
                    candidate_id = 0
            else:
                candidate_id = 0

    fragment.identity_corrected_solving_jumps = candidate_id


def compute_velocities_consecutive_fragments(
    neighbour_fragment_past: Fragment | None,
    fragment: Fragment,
    neighbour_fragment_future: Fragment | None,
) -> list[float]:
    """Compute velocities in the extremes of a `fragment` with respecto to its
    `neighbour_fragment_past` and `neighbour_fragment_future`

    Parameters
    ----------
    neighbour_fragment_past : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        past
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal
    neighbour_fragment_future : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        future

    Returns
    -------
    velocities : list
        List with the velocity needed to link the fragment with its fragment in
        the past and in the future

    See Also
    --------
    Fragment

    """
    velocities = [np.nan, np.nan]
    if neighbour_fragment_past is not None:
        velocities[0] = fragment.compute_border_velocity(neighbour_fragment_past)
    if neighbour_fragment_future is not None:
        velocities[1] = neighbour_fragment_future.compute_border_velocity(fragment)
    return velocities


def get_fragment_with_same_identity(
    number_of_frames: int,
    list_of_fragments: ListOfFragments,
    fragment: Fragment,
    direction: str,
) -> tuple[Fragment | None, int]:
    """Get the `neighbour_fragment` with the same identity in a given `direction`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving
        and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and
        methods related to them
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal
    direction : string
        If `direction` = `to_the_past` gets the `neighbour_fragment` in the
        past, `direction` = `to_the_future` gets the `neighbour_fragment` in
        the future

    Returns
    -------
    neighbour_fragment : <Fragment object>
        `Fragment` object with the same identity in a given `direction`
    number_of_frames_in_direction : int
        Number of frames to find the `neighbour_fragment` from a given extreme
        of the `fragment`

    See Also
    --------
    Fragment

    """
    number_of_frames_in_direction = 0
    frame_number = (
        fragment.start_frame if direction == "to_the_past" else fragment.end_frame
    )

    neighbour_fragment = None
    while neighbour_fragment is None and 0 < frame_number < number_of_frames:
        neighbour_fragment = fragment.get_neighbour_fragment(
            list_of_fragments.fragments,
            direction,
            number_of_frames_in_direction=number_of_frames_in_direction,
        )
        number_of_frames_in_direction += 1
        frame_number += -1 if direction == "to_the_past" else 1

    return neighbour_fragment, number_of_frames_in_direction


def compute_neighbour_fragments_and_velocities(
    number_of_frames: int, list_of_fragments: ListOfFragments, fragment: Fragment
) -> tuple[Fragment | None, Fragment | None, np.ndarray]:
    """Computes the fragments with the same identities to the past and to the
    future of a given `fragment` and gives the velocities at the extremes of
    the current `fragment`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving
        and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and
        methods related to them
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal

    Returns
    -------
    neighbour_fragment_past : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        past
    neighbour_fragment_future : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        future
    velocities_between_fragments : nd.array
        Velocities needed to connect the current fragment to its consecutive
        fragments in the past and in the future.

    See Also
    --------
    get_fragment_with_same_identity
    compute_velocities_consecutive_fragments

    """
    neighbour_fragment_past, n_frames_in_past = get_fragment_with_same_identity(
        number_of_frames, list_of_fragments, fragment, "to_the_past"
    )
    neighbour_fragment_future, n_frames_in_future = get_fragment_with_same_identity(
        number_of_frames, list_of_fragments, fragment, "to_the_future"
    )

    velocities = compute_velocities_consecutive_fragments(
        neighbour_fragment_past, fragment, neighbour_fragment_future
    )
    velocities_between_fragments = np.asarray(velocities) / np.asarray(
        [n_frames_in_past, n_frames_in_future]
    )

    return (
        neighbour_fragment_past,
        neighbour_fragment_future,
        velocities_between_fragments,
    )


def correct_impossible_velocity_jumps_loop(
    video: Video, list_of_fragments: ListOfFragments, scope: str
):
    """Checks whether the velocity needed to join two consecutive fragments with
    the same identity is consistent with the typical velocity of the animals in
    the video (`video.velocity_threshold`). If the velocity is not consistent the
    identity of one of the fragments is reassigned. The check is performed from the
    `video.first_frame_first_global_fragment` to the past or to the future according
    to the `scope`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving
        and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them
    scope : string
        If `scope` = `to_the_past` the check is performed to the past and if
        `scope` = `to_the_future` the check is performed to the future.

    See Also
    --------
    Fragment
    compute_neighbour_fragments_and_velocities
    compute_velocities_consecutive_fragments
    reassign

    """
    fragments_in_direction = list_of_fragments.get_ordered_list_of_fragments(
        scope, video.first_frame_first_global_fragment[video.accumulation_trial]
    )
    impossible_velocity_threshold = video.velocity_threshold

    for fragment in track(
        fragments_in_direction, f"Correcting impossible velocity jumps {scope}"
    ):
        # TODO This loops is too slow
        if fragment.is_a_crossing or fragment.assigned_identities[0] == 0:
            continue

        (
            neighbour_fragment_past,
            neighbour_fragment_future,
            velocities_between_fragments,
        ) = compute_neighbour_fragments_and_velocities(
            video.number_of_frames, list_of_fragments, fragment
        )

        if all(velocities_between_fragments > impossible_velocity_threshold):
            assert neighbour_fragment_future is not None
            assert neighbour_fragment_past is not None
            if (
                neighbour_fragment_past.identity_is_fixed
                or neighbour_fragment_future.identity_is_fixed
            ):
                reassign(fragment, list_of_fragments, impossible_velocity_threshold)
            else:
                neighbour_fragment_past_past = (
                    neighbour_fragment_past.get_neighbour_fragment(
                        list_of_fragments.fragments, "to_the_past"
                    )
                )
                velocity_in_past = compute_velocities_consecutive_fragments(
                    neighbour_fragment_past_past, neighbour_fragment_past, fragment
                )[0]
                neighbour_fragment_future_future = (
                    neighbour_fragment_future.get_neighbour_fragment(
                        list_of_fragments.fragments, "to_the_future"
                    )
                )
                velocity_in_future = compute_velocities_consecutive_fragments(
                    fragment,
                    neighbour_fragment_future,
                    neighbour_fragment_future_future,
                )[1]
                if (
                    velocity_in_past < impossible_velocity_threshold
                    or velocity_in_future < impossible_velocity_threshold
                ):
                    reassign(fragment, list_of_fragments, impossible_velocity_threshold)
        elif velocities_between_fragments[0] > impossible_velocity_threshold:
            assert neighbour_fragment_past is not None
            if neighbour_fragment_past.identity_is_fixed:
                reassign(fragment, list_of_fragments, impossible_velocity_threshold)
            else:
                reassign(
                    neighbour_fragment_past,
                    list_of_fragments,
                    impossible_velocity_threshold,
                )
        elif velocities_between_fragments[1] > impossible_velocity_threshold:
            assert neighbour_fragment_future is not None
            if neighbour_fragment_future.identity_is_fixed:
                reassign(fragment, list_of_fragments, impossible_velocity_threshold)
            else:
                reassign(
                    neighbour_fragment_future,
                    list_of_fragments,
                    impossible_velocity_threshold,
                )


def correct_impossible_velocity_jumps(video: Video, list_of_fragments: ListOfFragments):
    """Corrects the parts of the video where the velocity of any individual is
    higher than a particular velocity threshold given by `video.velocity_threshold`.
    This check is done from the `video.first_frame_first_global_fragment` to the
    past and to the future

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them

    See Also
    --------
    correct_impossible_velocity_jumps_loop

    """
    correct_impossible_velocity_jumps_loop(
        video, list_of_fragments, scope="to_the_past"
    )
    correct_impossible_velocity_jumps_loop(
        video, list_of_fragments, scope="to_the_future"
    )



from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import QLabel, QListWidget, QVBoxLayout, QWidget

from idtrackerai import Blob, Fragment
from idtrackerai_GUI_tools import key_event_modifier


class CustomListWidget(QListWidget):
    def __init__(self):
        super().__init__()
        self.setAlternatingRowColors(True)

    def keyPressEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyReleaseEvent(event)


class AdditionalInfo(QWidget):
    fragments: list[Fragment] | None

    def __init__(self) -> None:
        super().__init__()
        self.setLayout(QVBoxLayout())
        self.n_blobs_in_frame = QLabel("")
        self.blob_title = QLabel("Selected blob:")
        self.blob_properties = CustomListWidget()
        self.fragment_title = QLabel("Selected blob's fragment")
        self.fragment_properties = CustomListWidget()
        self.layout().setContentsMargins(0, 0, 0, 8)
        self.layout().addWidget(self.n_blobs_in_frame)
        self.layout().addWidget(self.blob_title)
        self.layout().addWidget(self.blob_properties)
        self.layout().addWidget(self.fragment_title)
        self.layout().addWidget(self.fragment_properties)

    def set_data(self, blob: Blob | None, n_blobs: int):
        self.blob_properties.clear()
        self.fragment_properties.clear()
        self.n_blobs_in_frame.setText(f"{n_blobs} blobs in frame")
        if blob is None:
            return

        try:
            self.blob_properties.addItems(blob.properties)
        except AttributeError:
            self.blob_properties.addItem("Corrupted Blob")

        if self.fragments is None:
            return

        try:
            self.fragment_properties.addItems(
                self.fragments[blob.fragment_identifier].properties
            )
        except AttributeError:
            self.fragment_properties.addItem("Corrupted Fragment")



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
import pickle
from pathlib import Path
from typing import Iterable

import numpy as np

from . import Blob, Fragment, GlobalFragment
from .utils import conf, resolve_path


class ListOfGlobalFragments:
    """Contains a list of instances of the class
    :class:`global_fragment.GlobalFragment`.

    It contains methods to retrieve information from these global fragments
    and to update their attributes.
    These methods are manily used during the cascade of training and
    identification protocols.

    Parameters
    ----------
    global_fragments : list
        List of instances of :class:`global_fragment.GlobalFragment`.
    """

    non_accumulable_global_fragments: list[GlobalFragment]
    """List of global fragments which are NOT candidate for accumulation"""

    global_fragments: list[GlobalFragment]
    """List of global fragments which are candidate for accumulation"""

    def __init__(self, global_fragments: Iterable[GlobalFragment]):
        self.global_fragments = []
        self.non_accumulable_global_fragments = []

        for global_fragment in global_fragments:
            if (
                global_fragment.min_n_images_per_fragment
                > conf.MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION
            ):
                self.global_fragments.append(global_fragment)
            else:
                self.non_accumulable_global_fragments.append(global_fragment)

        logging.info(
            "Total number of global_fragments: %d",
            len(self.non_accumulable_global_fragments) + len(self.global_fragments),
        )
        logging.info(
            "Of which %d are long enough to be accumulated", len(self.global_fragments)
        )

    @classmethod
    def from_fragments(
        cls,
        blobs_in_video: list[list[Blob]],
        fragments: list[Fragment],
        num_animals: int,
    ):
        """Creates the list of instances of the class
        :class:`~globalfragment.GlobalFragment`
        used to create :class:`.ListOfGlobalFragments`.

        Parameters
        ----------
        blobs_in_video : list
            List of lists with instances of the class class :class:`blob.Blob`).
        fragments : list
            List of instances of the class :class:`fragment.Fragment`
        num_animals : int
            Number of animals to be tracked as indicated by the user.

        Returns
        -------
        list
            list of instances of the class :class:`~globalfragment.GlobalFragment`

        """
        global_fragments_boolean_array = [
            is_global_fragment_core(blobs_in_frame, blobs_in_video[i - 1], num_animals)
            for i, blobs_in_frame in enumerate(blobs_in_video)
        ]

        indices_beginning_of_fragment = detect_global_fragments_core_first_frame(
            global_fragments_boolean_array
        )

        return cls(
            GlobalFragment(blobs_in_video, fragments, i)
            for i in indices_beginning_of_fragment
        )

    @property
    def single_global_fragment(self) -> bool:
        return len(self.global_fragments) == 1

    @property
    def no_global_fragment(self) -> bool:
        return len(self.global_fragments) == 0

    def sort_by_distance_travelled(self):
        self.global_fragments.sort(
            key=lambda x: x.minimum_distance_travelled, reverse=True
        )

    def order_by_distance_to_the_frame(self, frame_number: int):
        """Sorts the global fragments with respect to their distance from the
        first global fragment chose for accumulation.

        Parameters
        ----------
        video : :class:`video.Video`
            Instance of the class :class:`video.Video`.
        accumulation_trial : int
            accumulation number (protocol 2 performs a single accumulation
            attempt, and if used, protocol 3 will perform 3 other attempts)
        """
        self.global_fragments.sort(
            key=lambda x: abs(x.first_frame_of_the_core - frame_number)
        )

    def save(self, path: Path | str):
        """Saves an instance of the class.

        Before saving the instances of fragments associated to every global
        fragment are removed and reset them after saving. This
        prevents problems when pickling objects inside of objects.

        Parameters
        ----------
        global_fragments_path : str
            Path where the object will be stored
        """
        path = resolve_path(path)
        logging.info(f"Saving ListOfGlobalFragments at {path}")
        path.parent.mkdir(exist_ok=True)

        json.dump(self.__dict__, path.open("w"), cls=GlobalFragmentsEncoder, indent=4)

    @classmethod
    def load(
        cls, path: Path | str, fragments: list[Fragment] | None = None
    ) -> "ListOfGlobalFragments":
        """Loads an instance of the class saved with :meth:`save` and
        associates individual fragments to each global fragment by calling
        :meth:`~relink_fragments_to_global_fragments`

        Parameters
        ----------

        path_to_load : str
            Path where the object to be loaded is stored.
        fragments : list
            List of all the instances of the class :class:`fragment.Fragment`
            in the video.
        """
        path = resolve_path(path)
        logging.info(f"Loading ListOfGlobalFragments from {path}")

        if not path.is_file():  # <=5.1.3 compatibility
            if not path.with_suffix(".pickle").is_file():
                raise FileNotFoundError(path)
            pickle.load(path.with_suffix(".pickle").open("rb")).save(path)

        list_of_global_fragments = cls.__new__(cls)
        json_data = json.load(path.open("r"))

        list_of_global_fragments.global_fragments = [
            GlobalFragment.from_json(g_frag_data, fragments)
            for g_frag_data in json_data["global_fragments"]
        ]

        list_of_global_fragments.non_accumulable_global_fragments = [
            GlobalFragment.from_json(g_frag_data, fragments)
            for g_frag_data in json_data["non_accumulable_global_fragments"]
        ]

        return list_of_global_fragments


def detect_global_fragments_core_first_frame(boolean_array: list[bool]) -> list[int]:
    """Detects the frame where the core of a global fragment starts.

    A core of a global fragment is the part of the global fragment where all
    the individuals are visible, i.e. the number of animals in the frame equals
    the number of animals in the video :boolean_array: array with True
    where the number of animals in the frame equals the number of animals in
    the video.
    """
    if all(boolean_array):
        return [0]
    return [
        i
        for i in range(len(boolean_array))
        if (boolean_array[i] and not boolean_array[i - 1])
    ]  # boolean_array[0] is always False


def is_global_fragment_core(
    blobs_in_frame: list[Blob], blobs_in_frame_past: list[Blob], n_animals: int
) -> bool:
    """Return True if the set of fragments identifiers in the current frame
    is the same as in the previous frame, otherwise returns false
    """
    all_in_frame = len(blobs_in_frame) == n_animals

    same_fragment_identifiers = {b.fragment_identifier for b in blobs_in_frame} == {
        b.fragment_identifier for b in blobs_in_frame_past
    }
    not_all_were_in_frame = len(blobs_in_frame_past) != n_animals
    return all_in_frame and (same_fragment_identifiers or not_all_were_in_frame)


class GlobalFragmentsEncoder(json.JSONEncoder):
    """Json encoder to serialize Global Fragments with styled indentation"""

    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)

        if isinstance(obj, GlobalFragment):
            serial = obj.__dict__.copy()
            serial.pop("individual_fragments", None)  # remove connections

            serial["individual_fragments_identifiers"] = (  # without indentation
                f"NotString{json.dumps(obj.individual_fragments_identifiers)}"
            )

            return serial

        if isinstance(obj, np.integer):
            return int(obj)

        if isinstance(obj, np.floating):
            return float(obj)

        return super().default(obj)

    def iterencode(self, obj, **kwargs):
        for encoded in super().iterencode(obj, **kwargs):
            if encoded.startswith('"NotString'):
                # remove colons and "NotString"
                yield encoded[10:-1]
            else:
                yield encoded



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

import cv2
import numpy as np

from idtrackerai import Blob, Video


def compute_erosion_disk(blobs_in_video: list[list[Blob]]) -> int:
    min_frame_distance_transform = []
    for blobs_in_frame in blobs_in_video:
        if blobs_in_frame:
            min_frame_distance_transform.append(
                compute_min_frame_distance_transform(blobs_in_frame)
            )

    return np.ceil(np.nanmedian(min_frame_distance_transform)).astype(int)
    # return np.ceil(np.nanmedian([compute_min_frame_distance_transform(video, blobs_in_frame)
    #                              for blobs_in_frame in blobs_in_video
    #                              if len(blobs_in_frame) > 0])).astype(np.int)


def compute_min_frame_distance_transform(blobs_in_frame: list[Blob]) -> float:
    max_distance_transform = []
    for blob in blobs_in_frame:
        if blob.is_an_individual:
            try:
                max_distance_transform.append(
                    np.max(
                        cv2.distanceTransform(
                            blob.get_bbox_mask(), cv2.DIST_L2, cv2.DIST_MASK_PRECISE
                        )
                    )
                )
            except cv2.error:
                logging.warning("Could not compute distance transform for this blob")
    return np.min(max_distance_transform) if max_distance_transform else np.nan


def get_eroded_blobs(
    video: Video, blobs_in_frame: list[Blob], frame_number: int
) -> list[Blob]:
    segmented_frame = np.zeros((video.height, video.width), np.uint8)

    for blob in blobs_in_frame:
        segmented_frame = cv2.fillPoly(segmented_frame, (blob.contour,), 255)

    segmented_eroded_frame = cv2.erode(
        src=segmented_frame,
        kernel=np.ones(video.erosion_kernel_size, np.uint8),
        iterations=1,
    )

    # Extract blobs info
    contours = cv2.findContours(
        segmented_eroded_frame, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_KCOS
    )[0]

    return [
        Blob(contour, frame_number=frame_number, pixels_are_from_eroded_blob=True)
        for contour in contours
    ]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

from idtrackerai import ListOfBlobs


def assign_zeros_with_interpolation_identities(
    list_of_blobs: ListOfBlobs, list_of_blobs_no_gaps: ListOfBlobs
):
    logging.debug("creating copy of list_of_blobs")

    for blobs_in_frame, blobs_in_frame_no_gaps in zip(
        list_of_blobs.blobs_in_video, list_of_blobs_no_gaps.blobs_in_video
    ):
        unassigned_blobs = [
            blob
            for blob in blobs_in_frame
            if blob.is_an_individual and blob.assigned_identities[0] == 0
        ]
        for unassigned_blob in unassigned_blobs:
            candidate_blobs = [
                blob
                for blob in blobs_in_frame_no_gaps
                if blob.fragment_identifier == unassigned_blob.fragment_identifier
            ]
            if (
                len(candidate_blobs) == 1
                and len(candidate_blobs[0].assigned_identities) == 1
            ):
                unassigned_blob.identities_corrected_closing_gaps = candidate_blobs[
                    0
                ].assigned_identities

    return list_of_blobs



import logging

import numpy as np
from torch.nn import Module

from idtrackerai import GlobalFragment, Video
from idtrackerai.network import NetworkParams, fc_weights_reinit
from idtrackerai.utils import conf

from .accumulation_manager import (
    get_P1_array_and_argsort,
    p1_below_random,
    set_fragment_temporary_id,
)
from .assigner import compute_identification_statistics_for_non_accumulated_fragments
from .identity_network import get_predictions_identities


def identify_first_global_fragment_for_accumulation(
    first_global_fragment_for_accumulation: GlobalFragment,
    video: Video,
    identification_model: Module | None,
    network_params: NetworkParams,
):
    if (
        identification_model is not None and video.identity_transfer
    ):  # identity transfer
        logging.info(f"Transferring identities from {video.knowledge_transfer_folder}")
        identities = get_transferred_identities(
            first_global_fragment_for_accumulation,
            video,
            identification_model,
            network_params,
        )

        if identities is None:
            logging.warning(
                "[red bold]Identity transfer failed", extra={"markup": True}
            )
            logging.info(
                "We proceed by reinitializing fully connected layers, "
                "assigning random identities to the first GlobalFragment "
                "and transferring only the convolutional filters "
                "(knowledge transfer)"
            )
            identification_model.apply(fc_weights_reinit)
            identities = np.arange(video.n_animals)
        else:
            logging.info(
                "[green bold]Identities transferred successfully!",
                extra={"markup": True},
            )
    else:
        logging.info(
            "Tracking without identity transfer, assigning random initial identities"
        )
        identities = np.arange(video.n_animals)

    for id, fragment in zip(
        identities, first_global_fragment_for_accumulation.individual_fragments
    ):
        fragment.acceptable_for_training = True
        fragment.temporary_id = id
        frequencies = np.zeros(video.n_animals)
        frequencies[id] = fragment.number_of_images
        fragment.is_certain = True
        fragment.certainty = 1.0
        fragment.set_P1_from_frequencies(frequencies)


def get_transferred_identities(
    first_global_fragment_for_accumulation: GlobalFragment,
    video: Video,
    identification_model: Module,
    network_params: NetworkParams,
) -> list[int | None] | None:
    images, _ = first_global_fragment_for_accumulation.get_images_and_labels(
        video.id_images_file_paths
    )

    predictions, softmax_probs = get_predictions_identities(
        identification_model, images, network_params
    )

    compute_identification_statistics_for_non_accumulated_fragments(
        first_global_fragment_for_accumulation.individual_fragments,
        predictions,
        softmax_probs,
        network_params.n_classes,
    )

    # Check certainties of the individual fragments in the global fragment
    # for individual_fragment_identifier in global_fragment.individual_fragments_identifiers:

    for fragment in first_global_fragment_for_accumulation.individual_fragments:
        fragment.acceptable_for_training = True

    for fragment in first_global_fragment_for_accumulation.individual_fragments:
        if fragment.certainty < conf.CERTAINTY_THRESHOLD:
            logging.error(
                "A fragment is not certain enough, "
                f"CERTAINTY_THRESHOLD = {conf.CERTAINTY_THRESHOLD:.2f}, "
                f"fragment certainty = {fragment.certainty:.2f}"
            )
            return None

    P1_array, index_individual_fragments_sorted_by_P1 = get_P1_array_and_argsort(
        first_global_fragment_for_accumulation
    )

    # assign temporary identity to individual fragments by hierarchical P1
    for fragment_indx in index_individual_fragments_sorted_by_P1:
        fragment = first_global_fragment_for_accumulation.individual_fragments[
            fragment_indx
        ]

        if p1_below_random(P1_array, fragment_indx, fragment):
            logging.error("The computed identities P1 is below random")
            return None

        temporary_id = int(np.argmax(P1_array[fragment_indx, :]))
        if not fragment.check_consistency_with_coexistent_individual_fragments(
            temporary_id
        ):
            logging.error("The computed identities are not consistent")
            return None
        P1_array = set_fragment_temporary_id(
            fragment, temporary_id, P1_array, fragment_indx
        )

    # Check if the global fragment is unique after assigning the identities
    if not first_global_fragment_for_accumulation.is_unique(video.n_animals):
        logging.error("The computed identities are not unique")
        return None

    return [
        fragment.temporary_id
        for fragment in first_global_fragment_for_accumulation.individual_fragments
    ]



import logging

import cv2
import numpy as np

from idtrackerai import Video
from idtrackerai.utils import create_dir, track
from idtrackerai_GUI_tools import VideoPathHolder


# TODO INDIVIDUAL_VIDEO_WIDTH_HEIGHT parameter
def draw_general_frame(
    positions: list[tuple[int, int]],
    size: int,
    miniframes: np.ndarray,
    canvas: np.ndarray,
):
    for cur_id in range(miniframes.shape[0]):
        draw_x, draw_y = positions[cur_id]
        canvas[draw_y : draw_y + size, draw_x : draw_x + size] = miniframes[cur_id]


def read_individual_miniframes(
    frame: np.ndarray, ordered_centroid: np.ndarray, miniframes: np.ndarray
):
    if frame.ndim == 2:
        frame = frame[..., None]
    miniframes[:] = 0
    size2 = miniframes.shape[1] // 2
    for cur_id, (x, y) in enumerate(ordered_centroid):
        if x > 0 and y > 0:
            miniframe = frame[
                max(0, y - size2) : y + size2, max(0, x - size2) : x + size2
            ]
            miniframes[cur_id, 0 : miniframe.shape[0], 0 : miniframe.shape[1]] = (
                miniframe
            )


def generate_individual_video(
    video: Video,
    trajectories: np.ndarray,
    draw_in_gray: bool,
    starting_frame: int,
    ending_frame: int | None,
):
    if draw_in_gray:
        logging.info("Drawing original video in grayscale")

    trajectories = np.nan_to_num(trajectories, nan=-1).astype(int)

    create_dir(video.individual_videos_folder)

    n_rows = int(np.sqrt(video.n_animals))
    n_cols = int(video.n_animals / n_rows - 0.0001) + 1

    miniframe_size = 2 * (int(video.median_body_length_full_resolution) // 2)
    extra_lower_pad = 10
    bbox_side_pad = 10
    bbox_top_pad = 30
    full_bbox_width = miniframe_size + 2 * bbox_side_pad
    out_video_width = n_cols * full_bbox_width

    full_bbox_height = miniframe_size + bbox_top_pad
    out_video_height = n_rows * full_bbox_height + extra_lower_pad

    positions = [
        (
            full_bbox_width * (i % n_cols) + bbox_side_pad,
            full_bbox_height * (i // n_cols) + bbox_top_pad,
        )
        for i in range(video.n_animals)
    ]

    videoPathHolder = VideoPathHolder(video.video_paths)

    ending_frame = len(trajectories) - 1 if ending_frame is None else ending_frame
    logging.info(f"Drawing from frame {starting_frame} to {ending_frame}")

    general_video_writer = cv2.VideoWriter(
        str(video.individual_videos_folder / "general.avi"),
        cv2.VideoWriter_fourcc(*"XVID"),
        video.frames_per_second,
        (out_video_width, out_video_height),
    )

    individual_video_writers = [
        cv2.VideoWriter(
            str(video.individual_videos_folder / f"individual_{id+1}.avi"),
            cv2.VideoWriter_fourcc(*"XVID"),
            video.frames_per_second,
            (miniframe_size, miniframe_size),
        )
        for id in range(video.n_animals)
    ]

    labels = video.identities_labels or list(map(str, range(1, video.n_animals + 1)))

    miniframes = np.empty(
        (video.n_animals, miniframe_size, miniframe_size, 3), np.uint8
    )

    general_frame = np.zeros((out_video_height, out_video_width, 3), np.uint8)
    for cur_id in range(video.n_animals):
        draw_x, draw_y = positions[cur_id]
        general_frame = cv2.putText(
            general_frame,
            labels[cur_id],
            (draw_x, draw_y - 5),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.5,
            (255, 255, 255),
        )

    for frame in track(range(starting_frame, ending_frame), "Generating video"):
        img = videoPathHolder.read_frame(frame, not draw_in_gray)

        read_individual_miniframes(img, trajectories[frame], miniframes)

        draw_general_frame(positions, miniframe_size, miniframes, general_frame)

        general_video_writer.write(general_frame)

        for id in range(video.n_animals):
            individual_video_writers[id].write(miniframes[id])

    logging.info(f"Videos generated in {video.individual_videos_folder}")



from idtrackerai import ListOfBlobs, Video
from idtrackerai.utils import create_dir

from .crossing_detector import detect_crossings
from .model_area import compute_body_length


def crossings_detection_API(video: Video, list_of_blobs: ListOfBlobs) -> None:
    """
    This crossings detector works under the following assumptions
        1. The number of animals in the video is known (given by the user)
        2. There are frames in the video where all animals are separated from
        each other.
        3. All animals have a similar size
        4. The frame rate of the video is higher enough so that consecutive
        segmented blobs of pixels of the same animal overlap, i.e. some of the
        pixels representing the animal A in frame i are the same in the
        frame i+1.

    NOTE: This crossing detector sets the identification images that will be
    used to identify the animals
    """
    video.crossing_detector_timer.start()

    median_body_length = compute_body_length(list_of_blobs, video.n_animals)
    video.set_id_image_size(median_body_length)

    create_dir(video.id_images_folder, remove_existing=True)

    list_of_blobs.set_images_for_identification(
        video.episodes,
        video.id_images_file_paths,
        video.id_image_size,
        video.segmentation_data_folder,
        video.number_of_parallel_workers,
    )
    list_of_blobs.compute_overlapping_between_subsequent_frames()

    if video.single_animal:
        for blob in list_of_blobs.all_blobs:
            blob.is_an_individual = True
    else:
        detect_crossings(list_of_blobs, video)

    video.crossing_detector_timer.finish()


__all__ = ["crossings_detection_API"]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)

import logging
import os

import numpy as np


class GroundTruthBlob:
    """Lighter blob objects.
    Attributes:
        identity (preferring the one assigned by the user, if it is not None)
        centroid
        pixels (pixels is stored to check the groundtruth in crossings)
    """

    attributes = [
        "identity",
        "assigned_identity",
        "used_for_training",
        "accumulation_step",
        "centroid",
        "pixels",
        "frame_number",
        "is_an_individual",
        "is_a_crossing",
        "blob_index",
        "fragment_identifier",
    ]

    def get_attribute(self, blob):
        for attribute in self.attributes:
            if attribute == "identity":
                setattr(self, attribute, blob.final_identity)
            else:
                setattr(self, attribute, getattr(blob, attribute))


class IndividualGroundTruth:
    def __init__(
        self,
        video,
        individual_blobs_in_video,
        start=None,
        end=None,
        validated_identity=None,
    ):
        self.video = video
        self.individual_blobs_in_video = individual_blobs_in_video
        self.start = start
        self.end = end
        self.validated_identity = validated_identity

    def save(self):
        gt_name = "_individual_" + str(self.validated_identity) + "_groundtruth.npy"
        path_to_save_groundtruth = os.path.join(self.video.video_folder, gt_name)
        logging.info("saving ground truth at %s" % path_to_save_groundtruth)
        np.save(path_to_save_groundtruth, self)
        logging.info("done")


def generate_individual_groundtruth(
    video,
    blobs_in_video=None,
    start=None,
    end=None,
    validated_identity=None,
    save_gt=True,
):
    """Generates a list of light blobs_in_video, given a video object corresponding to a
    tracked video
    """
    individual_blobs_in_video_groundtruth = []

    for blobs_in_frame in blobs_in_video:
        for blob in blobs_in_frame:
            if blob.final_identity == validated_identity:
                gt_blob = GroundTruthBlob()
                gt_blob.get_attribute(blob)
                individual_blobs_in_video_groundtruth.append(gt_blob)

    groundtruth = IndividualGroundTruth(
        video=video,
        individual_blobs_in_video=individual_blobs_in_video_groundtruth,
        start=start,
        end=end,
        validated_identity=validated_identity,
    )
    if save_gt:
        groundtruth.save()
    return groundtruth



from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtWidgets import QFormLayout, QLineEdit, QScrollArea, QWidget


class IdLabels(QScrollArea):
    needToDraw = Signal()
    labels: list[str]

    def __init__(self):
        super().__init__()
        self.form_layout = QFormLayout()
        self.setWidgetResizable(True)
        wid = QWidget()
        wid.setLayout(self.form_layout)
        self.setWidget(wid)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def load_labels(self, labels: list[str]):
        for _ in range(self.form_layout.rowCount()):
            self.form_layout.removeRow(0)

        for indx, label in enumerate(labels, 1):
            edit = QLineEdit()
            edit.setText(label)
            edit.setPlaceholderText(str(indx))
            edit.setObjectName(str(indx))
            edit.textChanged.connect(self.new_label)
            edit.editingFinished.connect(self.validate_label)
            self.form_layout.addRow(f"{indx}:", edit)

        self.labels = [""] + labels

    def validate_label(self):
        sender = self.sender()
        assert isinstance(sender, QLineEdit)
        text = sender.text()
        if not text:
            sender.setText(sender.placeholderText())
        else:
            sender.setText(text.strip())

    def new_label(self, new_label=""):
        self.labels[int(self.sender().objectName())] = new_label
        self.needToDraw.emit()

    def get_labels(self):
        return self.labels



import logging
import sys
from contextlib import suppress

import numpy as np
import torch
from rich.console import Console
from rich.status import Status
from torch.backends import cudnn
from torch.nn import functional
from torch.utils.data import DataLoader

from idtrackerai.network import (
    DEVICE,
    LearnerClassification,
    NetworkParams,
    evaluate,
    train,
)
from idtrackerai.utils import CustomError, conf, track

from .identity_dataset import get_test_data_loader


class StopTraining:
    """Stops the training of the network according to the conditions specified
    in :meth:`__call__`

    Attributes
    ----------
    num_epochs : int
        Number of epochs before starting the training
    number_of_animals : int
        Number of animals in the video
    epochs_before_checking_stopping_conditions : int
        Number of epochs before starting to check the stopping conditions
    overfitting_counter : int
        Counts the number of consecutive overfitting epochs during training
    first_accumulation_flag : bool
        Flag to indicate that it is the first step of the accumulation

    """

    epochs_before_checking_stopping_conditions = 10

    def __init__(self, number_of_animals: int, is_first_accumulation: bool = False):
        logging.info("Setting the stopping criteria", stacklevel=3)
        self.num_epochs = conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN
        self.number_of_animals = number_of_animals
        # number of epochs in which the network is overfitting
        # before stopping the training
        self.overfitting_counter = 0
        self.is_first_accumulation = is_first_accumulation
        self.epochs_completed = -1

    def __call__(
        self,
        loss_training: float,
        loss_validation: list[float],
        accuracy_validation: float,
        status: Status,
    ):
        """Returns True when one of the conditions to stop the training is
        satisfied, otherwise it returns False"""
        self.epochs_completed += 1

        if self.epochs_completed > 0 and (
            np.isnan(loss_training) or np.isnan(loss_validation[-1])
        ):
            status.stop()
            logging.error(
                "The model diverged. Oops. Check the hyperparameters and "
                "the architecture of the network."
            )
            return True
        # check if it did not reached the epochs limit
        if self.epochs_completed >= self.num_epochs:
            status.stop()
            logging.warning(
                "The number of epochs completed is larger than the number "
                "of epochs set for training, we stop the training"
            )
            return True

        if self.epochs_completed <= self.epochs_before_checking_stopping_conditions:
            return False

        # check that the model is not overfitting or if it reached
        # a stable saddle (minimum)
        current_loss = loss_validation[-1]
        previous_loss = np.nanmean(
            loss_validation[-self.epochs_before_checking_stopping_conditions : -1]
        )

        # The validation loss in the first 10 epochs could have exploded
        # but being decreasing.
        if np.isnan(previous_loss):
            previous_loss = sys.float_info[0]
        losses_difference = previous_loss - current_loss

        # check overfitting
        if losses_difference < 0.0:
            self.overfitting_counter += 1
            if (
                not self.is_first_accumulation
                and self.overfitting_counter >= conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN
            ):
                status.stop()
                logging.info("Overfitting")
                return True
            if (
                self.is_first_accumulation
                and self.overfitting_counter
                > conf.OVERFITTING_COUNTER_THRESHOLD_IDCNN_FIRST_ACCUM
            ):
                status.stop()
                logging.info("Overfitting first accumulation")
                return True
        else:
            self.overfitting_counter = 0

        # check if the error is not decreasing much

        if abs(losses_difference) < conf.LEARNING_PERCENTAGE_DIFFERENCE_IDCNN * 10 ** (
            int(np.log10(current_loss)) - 1
        ):
            status.stop()
            logging.info("The losses difference is very small, we stop the training")
            return True

        # if the individual accuracies in validation are 1.
        # for all the animals
        if accuracy_validation == 1.0:
            status.stop()
            logging.info(
                "The individual accuracies in validation is 100% for "
                "all the individuals, we stop the training"
            )
            return True

        # if the validation loss is 0.
        if previous_loss == 0.0 or current_loss == 0.0:
            status.stop()
            logging.info("The validation loss is 0.0, we stop the training")
            return True

        return False


def TrainIdentification(
    learner: LearnerClassification,
    train_loader: DataLoader,
    val_loader: DataLoader,
    network_params: NetworkParams,
    stop_training: StopTraining,
):
    logging.info("Training Identification Network")

    # Initialize metric storage
    train_loss = 0.0
    val_loss = 0.0
    val_losses = []
    val_acc = 0.0

    logging.debug("Entering the epochs loop...")
    with Console().status("[red]Epochs loop...") as status:
        while not stop_training(train_loss, val_losses, val_acc, status):
            epoch = stop_training.epochs_completed

            train_loss = train(epoch, train_loader, learner)
            val_loss, val_acc = evaluate(val_loader, network_params.n_classes, learner)

            val_losses.append(val_loss)

            with suppress(IndexError):
                status.update(
                    f"[red]Epoch {epoch}: training loss = {train_loss:.6f},"
                    f" validation loss = {val_loss:.6f} and accuracy ="
                    f" {val_acc:.4%}"
                )

        logging.info("Last epoch loop: %s", status.status, extra={"markup": True})

    learner.save_model(network_params.model_path, val_acc=val_acc)

    if np.isnan(train_loss) or np.isnan(val_loss):
        raise CustomError("The model diverged")

    logging.info("Identification network trained")


def get_predictions_identities(
    model: torch.nn.Module, images: np.ndarray, network_params: NetworkParams
):
    logging.debug("Generating prediction data set with %d images", len(images))
    loader = get_test_data_loader({"images": images}, network_params.n_classes)
    predictions = []
    softmax_probs = []

    logging.debug("Using trained network to predict images identities")
    if not next(model.parameters()).is_cuda:
        logging.info("Sending model and criterion to %s", DEVICE)
        cudnn.benchmark = True  # make it train faster
        model.to(DEVICE)

    model.eval()
    with torch.no_grad():
        for input, _target in track(loader, "Predicting identities"):
            # Inference
            softmax = functional.softmax(model.forward(input.to(DEVICE)), dim=1)
            pred = softmax.argmax(1)  # find the predicted class

            predictions += pred.tolist()
            softmax_probs += softmax.tolist()

    return np.asarray(predictions) + 1, np.asarray(softmax_probs)



import logging
from typing import Iterable

import numpy as np

from idtrackerai import Blob, ListOfBlobs, ListOfFragments, Video
from idtrackerai.utils import create_dir

from .assign_them_all import close_trajectories_gaps
from .compute_velocity_model import compute_model_velocity
from .correct_impossible_velocity_jumps import correct_impossible_velocity_jumps
from .get_trajectories import produce_output_dict
from .identify_non_assigned_with_interpolation import (
    assign_zeros_with_interpolation_identities,
)
from .trajectories_to_csv import convert_trajectories_file_to_csv_and_json


def trajectories_API(
    video: Video,
    list_of_blobs: ListOfBlobs,
    single_global_fragment: bool,
    list_of_fragments: ListOfFragments,
):
    if (
        not video.track_wo_identities
        and not video.single_animal
        and not single_global_fragment
    ):
        postprocess_impossible_jumps(video, list_of_fragments, list_of_blobs.all_blobs)

    video.create_trajectories_timer.start()
    create_dir(video.trajectories_folder, remove_existing=True)

    trajectories = produce_output_dict(
        list_of_blobs.blobs_in_video, video, list_of_fragments.fragments
    )

    trajectories_file = video.trajectories_folder / "with_gaps.npy"

    logging.info(f"Saving trajectories with gaps in {trajectories_file}")
    np.save(trajectories_file, trajectories)  # type: ignore
    if video.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_file, video.add_time_column_to_csv
        )

    list_of_blobs.save(video.blobs_path)
    del list_of_blobs
    if (
        not video.track_wo_identities
        and not video.single_animal
        and not single_global_fragment
    ):
        interpolate_crossings(video, list_of_fragments)
    else:
        video.estimated_accuracy = 1.0
    video.create_trajectories_timer.finish()
    video.general_timer.finish()
    video.save()


def postprocess_impossible_jumps(
    video: Video, list_of_fragments: ListOfFragments, all_blobs: Iterable[Blob]
):
    video.impossible_jumps_timer.start()
    video.velocity_threshold = compute_model_velocity(list_of_fragments)
    correct_impossible_velocity_jumps(video, list_of_fragments)

    video.individual_fragments_stats = list_of_fragments.get_stats()

    video.estimated_accuracy = compute_estimated_accuracy(list_of_fragments)
    list_of_fragments.save(video.accumulation_folder / "list_of_fragments.json")
    list_of_fragments.update_blobs(all_blobs)
    video.impossible_jumps_timer.finish()


def compute_estimated_accuracy(list_of_fragments: ListOfFragments) -> float:
    weighted_P2 = 0
    number_of_individual_blobs = 0

    for fragment in list_of_fragments.individual_fragments:
        if fragment.assigned_identities[0] not in (0, None):
            assert fragment.P2_vector is not None
            weighted_P2 += (
                fragment.P2_vector[fragment.assigned_identities[0] - 1]
                * fragment.number_of_images
            )
        number_of_individual_blobs += fragment.number_of_images
    return weighted_P2 / number_of_individual_blobs


def interpolate_crossings(video: Video, list_of_fragments: ListOfFragments):
    video.crossing_solver_timer.start()
    list_of_blobs_no_gaps = close_trajectories_gaps(
        video, ListOfBlobs.load(video.blobs_path), list_of_fragments
    )
    list_of_blobs_no_gaps.save(video.blobs_no_gaps_path)
    video.crossing_solver_timer.finish()

    trajectories_wo_gaps_file = video.trajectories_folder / "without_gaps.npy"
    logging.info(
        "Generating trajectories. The trajectories files are stored in "
        f"{trajectories_wo_gaps_file}"
    )
    trajectories_wo_gaps = produce_output_dict(
        list_of_blobs_no_gaps.blobs_in_video, video, list_of_fragments.fragments
    )

    np.save(trajectories_wo_gaps_file, trajectories_wo_gaps)  # type: ignore
    if video.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_wo_gaps_file, video.add_time_column_to_csv
        )

    # Now, two ListOfBlobs will be loaded in RAM, we clean the heavier parts of
    # the objects to free space. These light versions of ListOfBlobs
    # should not be saved
    for blob in list_of_blobs_no_gaps.all_blobs:
        del blob.contour
        if hasattr(blob, "convexHull"):
            del blob.convexHull

    list_of_blobs = ListOfBlobs.load(video.blobs_path)
    for blob in list_of_blobs.all_blobs:
        del blob.contour
        if hasattr(blob, "convexHull"):
            del blob.convexHull

    logging.info("Saving trajectories")
    list_of_blobs = assign_zeros_with_interpolation_identities(
        list_of_blobs, list_of_blobs_no_gaps
    )
    trajectories_file = video.trajectories_folder / "with_gaps.npy"
    trajectories = produce_output_dict(
        list_of_blobs.blobs_in_video, video, list_of_fragments.fragments
    )
    np.save(trajectories_file, trajectories)  # type: ignore
    if video.convert_trajectories_to_csv_and_json:
        convert_trajectories_file_to_csv_and_json(
            trajectories_file, video.add_time_column_to_csv
        )



from importlib import metadata

# Video has to be the first class to be imported
from idtrackerai.video import Video

from .blob import Blob
from .fragment import Fragment
from .globalfragment import GlobalFragment
from .list_of_blobs import ListOfBlobs
from .list_of_fragments import ListOfFragments
from .list_of_global_fragments import ListOfGlobalFragments

__version__ = metadata.version("idtrackerai")


__all__ = [
    "Blob",
    "ListOfBlobs",
    "ListOfFragments",
    "ListOfGlobalFragments",
    "ListOfGlobalFragments",
    "GlobalFragment",
    "Video",
    "Fragment",
]



import logging

import torch
from torch.backends import cudnn
from torch.nn import CrossEntropyLoss
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import ListOfBlobs, Video
from idtrackerai.network import (
    DEVICE,
    LearnerClassification,
    NetworkParams,
    weights_xavier_init,
)
from idtrackerai.utils import conf

from .crossings_dataset import (
    get_train_validation_and_eval_blobs,
    get_training_data_loaders,
)
from .crossings_network import (
    StopTraining,
    get_predictions_crossigns,
    train_deep_crossing,
)
from .model_area import ModelArea


def apply_area_and_unicity_heuristics(list_of_blobs: ListOfBlobs, n_animals: int):
    logging.info(
        "Classifying Blobs as individuals or crossings "
        "depending on their area and the number of blobs in the frame"
    )

    model_area = ModelArea(list_of_blobs, n_animals)

    for blobs_in_frame in list_of_blobs.blobs_in_video:
        unicity_cond = len(blobs_in_frame) == n_animals
        for blob in blobs_in_frame:
            blob.seems_like_individual = unicity_cond or model_area(blob.area)


def detect_crossings(list_of_blobs: ListOfBlobs, video: Video):
    """Classify all blobs in the video as being crossings or individuals"""

    apply_area_and_unicity_heuristics(list_of_blobs, video.n_animals)

    train_blobs, val_blobs, eval_blobs = get_train_validation_and_eval_blobs(
        list_of_blobs.blobs_in_video, video.n_animals
    )

    if (
        len(train_blobs["crossings"])
        < conf.MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR
    ):
        logging.debug("There are not enough crossings to train the crossing detector")
        for blob in eval_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return
    logging.info("There are enough crossings to train the crossing detector")
    train_loader, val_loader = get_training_data_loaders(
        video.id_images_file_paths, train_blobs, val_blobs
    )
    logging.info("Setting crossing detector network parameters")
    network_params = NetworkParams(
        n_classes=2,
        architecture="DCD",
        save_folder=video.crossings_detector_folder,
        model_name="crossing_detector",
        image_size=video.id_image_size,
        optimizer="Adam",
        schedule=[30, 60],
        optim_args={"lr": conf.LEARNING_RATE_DCD},
        epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_DCD,
    )
    network_params.save()

    criterion = CrossEntropyLoss(weight=torch.tensor(train_blobs["weights"]))
    crossing_detector_model = LearnerClassification.create_model(network_params)
    logging.info("Initialize networks params with Xavier initialization")
    crossing_detector_model.apply(weights_xavier_init)

    logging.info("Sending model and criterion to %s", DEVICE)
    cudnn.benchmark = True  # make it train faster
    crossing_detector_model.to(DEVICE)
    criterion.to(DEVICE)

    logging.info(f"Setting {network_params.optimizer} optimizer")
    if network_params.optimizer == "Adam":
        optimizer = torch.optim.Adam(
            crossing_detector_model.parameters(), **network_params.optim_args
        )
    elif network_params.optimizer == "SGD":
        optimizer = torch.optim.SGD(
            crossing_detector_model.parameters(), **network_params.optim_args
        )
    else:
        raise AttributeError(network_params.optimizer)

    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    learner = LearnerClassification(
        crossing_detector_model, criterion, optimizer, scheduler
    )

    stop_training = StopTraining(network_params.epochs)

    model_diverged, best_model_path = train_deep_crossing(
        learner, train_loader, val_loader, network_params, stop_training
    )

    if model_diverged:
        logging.warning(
            "[red]The model diverged[/] provably due to a bad segmentation. Falling"
            " back to individual-crossing discrimination by average area model.",
            extra={"markup": True},
        )
        for blob in eval_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return

    del train_loader
    del val_loader

    model_state = torch.load(best_model_path)
    crossing_detector_model.load_state_dict(model_state, strict=True)
    logging.info("Loaded best model weights from %s", best_model_path)

    logging.info("Using crossing detector to classify individuals and crossings")
    predictions = get_predictions_crossigns(
        video.id_images_file_paths, crossing_detector_model, eval_blobs
    )

    logging.info(
        "Prediction results: %d individuals and %d crossings",
        predictions.count(0),
        predictions.count(1),
    )
    for blob, prediction in zip(eval_blobs, predictions):
        blob.is_an_individual = prediction != 1

    list_of_blobs.update_id_image_dataset_with_crossings(video.id_images_file_paths)



import logging
from itertools import pairwise

import cv2
import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QImage, QPainter

from idtrackerai import Video
from idtrackerai.utils import track
from idtrackerai_GUI_tools import VideoPathHolder, get_cmap


def QImageToArray(qimg: QImage) -> np.ndarray:
    width = qimg.width()
    height = qimg.height()
    byte_str = qimg.bits()
    return np.frombuffer(byte_str.asarray(height * width * 4), np.uint8).reshape(
        (height, width, 4)
    )[:, :, :-1]


def setColormap(n_animals: int):
    return get_cmap()[[int(i * 255 / n_animals) for i in range(n_animals)]]


def draw_general_frame(
    np_frame: np.ndarray,
    frame_number: int,
    trajectories: np.ndarray,
    centroid_trace_length: int,
    colors: list[tuple[int, int, int]],
    labels: list[str],
) -> np.ndarray:
    ordered_centroid = trajectories[frame_number]
    frame = QImage(
        np_frame.data,
        np_frame.shape[1],
        np_frame.shape[0],
        np_frame.shape[1] * 3,
        QImage.Format.Format_BGR888,
    )
    canvas = QImage(frame.size(), QImage.Format.Format_ARGB32_Premultiplied)
    canvas.fill(Qt.GlobalColor.transparent)
    painter = QPainter(canvas)
    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
    pen = painter.pen()
    pen.setWidth(2)
    for cur_id, centroid in enumerate(ordered_centroid):
        if frame_number > centroid_trace_length:
            centroids_trace = trajectories[
                frame_number - centroid_trace_length : frame_number + 1, cur_id
            ]
        else:
            centroids_trace = trajectories[: frame_number + 1, cur_id]
        color = QColor(*colors[cur_id])

        alphas = np.linspace(0, 255, len(centroids_trace), dtype=int)[1:]
        if len(centroids_trace) > 1:
            for alpha, (pointA, pointB) in zip(alphas, pairwise(centroids_trace)):
                if any(pointA < 0) or any(pointB < 0):
                    continue
                color.setAlpha(alpha)
                pen.setColor(color)
                painter.setPen(pen)
                painter.drawLine(*pointA, *pointB)

        if all(centroid > 0):
            color.setAlpha(255)
            painter.setBrush(color)
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawEllipse(centroid[0] - 3, centroid[1] - 3, 6, 6)

    painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationOver)
    painter.drawImage(canvas.rect(), frame)
    painter.end()

    arr_img = np.array(QImageToArray(canvas))
    for cur_id, centroid in enumerate(ordered_centroid):
        if all(centroid > 0):
            color = (
                int(colors[cur_id][2]),
                int(colors[cur_id][1]),
                int(colors[cur_id][0]),
            )  # BGR

            arr_img = cv2.putText(
                arr_img,
                labels[cur_id],
                (centroid[0], centroid[1]),
                cv2.FONT_HERSHEY_COMPLEX,
                0.8,
                color,
                2,
            )

    return arr_img


def generate_trajectories_video(
    video: Video,
    trajectories: np.ndarray,
    draw_in_gray: bool,
    centroid_trace_length: int,
    starting_frame: int,
    ending_frame: int,
):
    if draw_in_gray:
        logging.info("Drawing original video in grayscale")

    resize_factor = min(1920 / video.original_width, 1080 / video.original_height, 1)

    if resize_factor != 1:
        logging.info(f"Applying resize of factor {resize_factor}")

    trajectories = np.nan_to_num(trajectories * resize_factor, nan=-1).astype(int)

    video_name = video.video_paths[0].stem + "_tracked.avi"

    colors = setColormap(video.n_animals)

    labels = video.identities_labels or list(map(str, range(1, video.n_animals + 1)))

    path_to_save_video = video.session_folder / video_name

    out_video_width = int(video.original_width * resize_factor)
    out_video_height = int(video.original_height * resize_factor)

    video_writer = cv2.VideoWriter(
        str(path_to_save_video),
        cv2.VideoWriter_fourcc(*"XVID"),
        video.frames_per_second,
        (out_video_width, out_video_height),
    )

    videoPathHolder = VideoPathHolder(video.video_paths)

    ending_frame = len(trajectories) - 1 if ending_frame is None else ending_frame
    logging.info(f"Drawing from frame {starting_frame} to {ending_frame}")

    for frame in track(range(starting_frame, ending_frame), "Generating video"):
        img = videoPathHolder.read_frame(frame, True)

        if resize_factor != 1:
            img = cv2.resize(img, (0, 0), fx=resize_factor, fy=resize_factor)

        if draw_in_gray:
            img = cv2.cvtColor(
                cv2.cvtColor(img, cv2.COLOR_BGR2GRAY), cv2.COLOR_GRAY2BGR
            )

        img = draw_general_frame(
            img, frame, trajectories, centroid_trace_length, colors, labels
        )

        video_writer.write(img)

    logging.info(f"Video generated in {path_to_save_video}")



"""This file provides the template Learner. The Learner is used in training/evaluation loop
The Learner implements the training procedure for specific task.
The default Learner is from classification task."""
import logging
from pathlib import Path

import torch
from torch.nn import CrossEntropyLoss, Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import MultiStepLR

from . import NetworkParams, models


class LearnerClassification:
    def __init__(
        self,
        model: Module,
        criterion: CrossEntropyLoss,
        optimizer: Optimizer,
        scheduler: MultiStepLR,
    ):
        super().__init__()
        logging.info("Setting the learner")
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.epoch: int = 0

    @staticmethod
    def create_model(learner_params: NetworkParams) -> Module:
        architecture = learner_params.architecture
        logging.info("Creating %s model", architecture)
        if architecture == "DCD":
            model = models.DCD
        elif architecture == "idCNN":
            model = models.idCNN
        elif architecture == "idCNN_adaptive":
            model = models.idCNN_adaptive
        else:
            raise ValueError(architecture)

        return model(
            out_dim=learner_params.n_classes, input_shape=learner_params.image_size
        )

    @classmethod
    def load_model(
        cls, learner_params: NetworkParams, knowledge_transfer: bool = False
    ):
        model = cls.create_model(learner_params)
        if knowledge_transfer:
            model_path = learner_params.knowledge_transfer_model_file
            assert model_path is not None
        else:
            model_path = learner_params.load_model_path

        logging.info("Load model weights from %s", model_path)
        # The path to model file (*.best_model.pth). Do NOT use checkpoint file here
        model_state: dict = torch.load(model_path)
        model_state.pop("val_acc", None)

        try:
            model.load_state_dict(model_state, strict=True)
        except RuntimeError:
            logging.warning(
                "Loading a model from a version older than 5.1.7, "
                "going to translate the state dictionary."
            )
            translated_model_state = {
                "layers.0.weight": model_state["conv1.weight"],
                "layers.0.bias": model_state["conv1.bias"],
                "layers.3.weight": model_state["conv2.weight"],
                "layers.3.bias": model_state["conv2.bias"],
                "layers.6.weight": model_state["conv3.weight"],
                "layers.6.bias": model_state["conv3.bias"],
                "layers.9.weight": model_state["fc1.weight"],
                "layers.9.bias": model_state["fc1.bias"],
                "layers.11.weight": model_state["fc2.weight"],
                "layers.11.bias": model_state["fc2.bias"],
            }
            model.load_state_dict(translated_model_state, strict=True)

        return model

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def forward_with_criterion(
        self, inputs: torch.Tensor, targets: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        out = self.model.forward(inputs)
        return self.criterion(out, targets), out

    def learn(self, inputs, targets):
        loss, out = self.forward_with_criterion(inputs, targets)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss

    def step_schedule(self, epoch):
        self.epoch = epoch
        self.scheduler.step()

    def save_model(self, savename: Path, **extra_data):
        logging.info("Saving model at %s", savename)
        torch.save(self.model.state_dict() | extra_data, savename)



from .additional_info import AdditionalInfo
from .errors_explorer import ErrorsExplorer
from .id_groups import IdGroups
from .id_labels import IdLabels
from .interpolator import Interpolator
from .mark_properties import MarkBlobs
from .paint_blobs import find_selected_blob, paintBlobs, paintTrails
from .setup_points import SetupPoints

__all__ = [
    "MarkBlobs",
    "paintBlobs",
    "IdGroups",
    "find_selected_blob",
    "IdLabels",
    "paintTrails",
    "ErrorsExplorer",
    "SetupPoints",
    "Interpolator",
    "AdditionalInfo",
]
